# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.utils.api_options import APIOptions, defaultAPIOptions


class TestAPIOptions:
    @pytest.mark.describe("test of header inheritance in APIOptions")
    def test_apioptions_headers(self) -> None:
        opts_d = defaultAPIOptions(environment="dev")
        opts_1 = opts_d.with_override(
            APIOptions(
                database_additional_headers={"d": "y", "D": None},
                admin_additional_headers={"a": "y", "A": None},
                redacted_header_names={"x", "y"},
            )
        )
        opts_2 = opts_d.with_override(
            APIOptions(
                database_additional_headers={"D": "y"},
                admin_additional_headers={"A": "y"},
                redacted_header_names={"x"},
            )
        ).with_override(
            APIOptions(
                database_additional_headers={"d": "y", "D": None},
                admin_additional_headers={"a": "y", "A": None},
                redacted_header_names={"y"},
            )
        )

        assert opts_1 == opts_2



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import (
    AstraDBAdmin,
    AstraDBDatabaseAdmin,
    AsyncDatabase,
    DataAPIClient,
    DataAPIDatabaseAdmin,
    Database,
)
from astrapy.authentication import (
    StaticTokenProvider,
    UsernamePasswordTokenProvider,
)
from astrapy.constants import Environment
from astrapy.exceptions import InvalidEnvironmentException
from astrapy.utils.api_options import (
    APIOptions,
    DataAPIURLOptions,
    DevOpsAPIURLOptions,
    defaultAPIOptions,
)

api_ep0123_dev = (
    "https://01234567-89ab-cdef-0123-456789abcdef-region.apps.astra-dev.datastax.com"
)
api_ep7777_dev = (
    "https://77777777-89ab-cdef-0123-456789abcdef-region.apps.astra-dev.datastax.com"
)
api_ep9999_test = (
    "https://99999999-89ab-cdef-0123-456789abcdef-region.apps.astra-test.datastax.com"
)


class TestAdminConversions:
    @pytest.mark.describe("test of DataAPIClient conversions and comparison functions")
    def test_dataapiclient_conversions(self) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        callers1 = [("x", "y")]
        dac1 = DataAPIClient(
            "t1",
            environment="dev",
            callers=callers0,
        )
        dac2 = DataAPIClient(
            "t1",
            environment="dev",
            callers=callers0,
        )

        assert dac1 == dac2
        assert dac1 != dac1._copy(token="x")
        assert dac1 == dac1._copy(token="x")._copy(token="t1")
        assert dac1 == dac1.with_options(token="x").with_options(
            api_options=APIOptions(token="t1")
        )
        assert dac1 == dac1.with_options(token="x").with_options(
            token="t1", api_options=APIOptions(token="t_another")
        )

        a_e_string = (
            "https://01234567-89ab-cdef-0123-456789abcdef-us-east1"
            ".apps.astra-dev.datastax.com"
        )
        db1 = dac1[a_e_string]
        opts0 = defaultAPIOptions(environment="dev").with_override(
            APIOptions(
                token="t1",
                callers=callers0,
            )
        )
        expected_db_1 = Database(
            api_endpoint=a_e_string,
            keyspace=None,
            api_options=opts0,
        )

        assert db1 == expected_db_1
        a_e_string_test = (
            "https://99999999-89ab-cdef-0123-456789abcdef-us-east1"
            ".apps.astra-test.datastax.com"
        )
        with pytest.raises(InvalidEnvironmentException):
            dac1[a_e_string_test]

        # equivalence between passing api_options and named parameters; option override
        dac1_opt = DataAPIClient(
            environment="dev",
            api_options=APIOptions(
                token="t1",
                callers=callers0,
            ),
        )
        dac1_opt_ovr = DataAPIClient(
            "t1",
            environment="dev",
            callers=callers0,
            api_options=APIOptions(
                token="t_another",
                callers=callers1,
            ),
        )
        assert dac1 == dac1_opt
        assert dac1 == dac1_opt_ovr

    @pytest.mark.describe("test of spawning databases from a DataAPIClient")
    def test_dataapiclient_spawning_databases(self) -> None:
        token = "the-token"
        database_id = "00000000-1111-2222-3333-444444444444"
        database_region = "the-region"
        endpoint = f"https://{database_id}-{database_region}.apps.astra.datastax.com"
        client = DataAPIClient(
            token=token,
            environment=Environment.PROD,
            callers=[("cn", "cv")],
        )
        db1 = client.get_database(endpoint)
        db2 = client.get_database(endpoint)
        assert db1 == db2

        # api_options and override check
        db2_param = client.get_database(endpoint, token="t1")
        db2_opt = client.get_database(
            endpoint,
            spawn_api_options=APIOptions(
                token="t1",
            ),
        )
        db2_opt_override = client.get_database(
            endpoint,
            token="t1",
            spawn_api_options=APIOptions(
                token="t_another",
            ),
        )
        assert db2_param == db2_opt
        assert db2_param == db2_opt_override

    @pytest.mark.describe("test of AstraDBAdmin conversions and comparison functions")
    def test_astradbadmin_conversions(self) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        opts0 = defaultAPIOptions(environment="dev").with_override(
            APIOptions(
                token="t1",
                callers=callers0,
                dev_ops_api_url_options=DevOpsAPIURLOptions(
                    dev_ops_url="dou",
                    dev_ops_api_version="dvv",
                ),
            ),
        )
        adm0 = AstraDBAdmin(api_options=opts0)
        adm1 = AstraDBAdmin(api_options=opts0)

        assert adm0 == adm1
        assert adm0 != adm0._copy(token="x")

        assert adm0 == adm0._copy(token="x")._copy(token="t1")

        assert adm0 != adm0.with_options(token="x")
        assert adm0 == adm0.with_options(token="x").with_options(token="t1")

        assert adm0 == adm0.with_options(token="x").with_options(
            api_options=APIOptions(token="t1")
        )
        assert adm0 == adm0.with_options(token="x").with_options(
            token="t1", api_options=APIOptions(token="t_another")
        )

    @pytest.mark.describe(
        "test of AstraDBDatabaseAdmin conversions and comparison functions"
    )
    def test_astradbdatabaseadmin_conversions(self) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        opts0 = defaultAPIOptions(environment="dev").with_override(
            APIOptions(
                token="t1",
                callers=callers0,
                dev_ops_api_url_options=DevOpsAPIURLOptions(
                    dev_ops_url="dou",
                    dev_ops_api_version="dvv",
                ),
                data_api_url_options=DataAPIURLOptions(
                    api_path="appi",
                    api_version="vX",
                ),
            ),
        )
        adda1 = AstraDBDatabaseAdmin(
            api_endpoint=api_ep0123_dev,
            api_options=opts0,
        )
        adda2 = AstraDBDatabaseAdmin(
            api_endpoint=api_ep0123_dev,
            api_options=opts0,
        )

        assert adda1 == adda2
        assert adda1 != adda1._copy(token="x")

        assert adda1 == adda1._copy(token="x")._copy(token="t1")
        assert adda1 != adda1.with_options(token="x")

        assert adda1 == adda1.with_options(token="x").with_options(token="t1")

        assert adda1 == adda1.with_options(token="x").with_options(
            api_options=APIOptions(token="t1")
        )
        assert adda1 == adda1.with_options(token="x").with_options(
            token="t1", api_options=APIOptions(token="t1_another")
        )

    @pytest.mark.describe(
        "test of DataAPIDBDatabaseAdmin conversions and comparison functions"
    )
    def test_dataapidatabaseadmin_conversions(self) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        opts0 = defaultAPIOptions(environment="hcd").with_override(
            APIOptions(
                token="t1",
                callers=callers0,
                data_api_url_options=DataAPIURLOptions(
                    api_path="appi",
                    api_version="v9",
                ),
            ),
        )
        dada1 = DataAPIDatabaseAdmin(
            api_endpoint="http://a.b.c:1234",
            api_options=opts0,
        )
        dada2 = DataAPIDatabaseAdmin(
            api_endpoint="http://a.b.c:1234",
            api_options=opts0,
        )
        assert dada1 == dada2

        assert dada1 != dada1._copy(token="tx")

        assert dada1 == dada1._copy(token="x")._copy(token="t1")

        assert dada1 != dada1.with_options(token="x")

        assert dada1 == dada1.with_options(token="x").with_options(token="t1")

        assert dada1 == dada1.with_options(token="x").with_options(
            api_options=APIOptions(token="t1")
        )
        assert dada1 == dada1.with_options(token="x").with_options(
            token="t1", api_options=APIOptions(token="t1_another")
        )

    @pytest.mark.describe("test of token inheritance in spawning from DataAPIClient")
    def test_dataapiclient_token_inheritance(self) -> None:
        client_t = DataAPIClient(token=StaticTokenProvider("static"))
        client_0 = DataAPIClient()
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        client_f = DataAPIClient(token=token_f)
        a_e_string = (
            "https://01234567-89ab-cdef-0123-456789abcdef-us-east1"
            ".apps.astra.datastax.com"
        )

        assert client_t.get_database(
            a_e_string, token=token_f
        ) == client_f.get_database(a_e_string)
        assert client_0.get_database(
            a_e_string, token=token_f
        ) == client_f.get_database(a_e_string)

        assert client_t.get_database(
            a_e_string, token=token_f
        ) == client_f.get_database(a_e_string)
        assert client_0.get_database(
            a_e_string, token=token_f
        ) == client_f.get_database(a_e_string)

        assert client_t.get_admin(token=token_f) == client_f.get_admin()
        assert client_0.get_admin(token=token_f) == client_f.get_admin()

    @pytest.mark.describe("test of client.get_admin option passing")
    def test_client_get_admin_option_passing(self) -> None:
        client_0 = DataAPIClient()
        admin_opt = client_0.get_admin(
            spawn_api_options=APIOptions(
                token="tx",
            ),
        )
        admin_param = client_0.get_admin(token="tx")
        admin_opt_param = client_0.get_admin(
            token="tx",
            spawn_api_options=APIOptions(
                token="t_another",
            ),
        )

        assert admin_opt == admin_param
        assert admin_opt == admin_opt_param

    @pytest.mark.describe("test of token inheritance in spawning from AstraDBAdmin")
    def test_astradbadmin_token_inheritance(self) -> None:
        opts_0 = defaultAPIOptions(environment=Environment.PROD)
        opts_t = opts_0.with_override(
            APIOptions(token=StaticTokenProvider("static")),
        )
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        opts_f = opts_0.with_override(APIOptions(token=token_f))
        admin_t = AstraDBAdmin(api_options=opts_t)
        admin_0 = AstraDBAdmin(api_options=opts_0)
        admin_f = AstraDBAdmin(api_options=opts_f)
        db_id_string = "01234567-89ab-cdef-0123-456789abcdef"

        assert admin_t.get_database(
            db_id_string, token=token_f, keyspace="n", region="r"
        ) == admin_f.get_database(db_id_string, keyspace="n", region="r")
        assert admin_0.get_database(
            db_id_string, token=token_f, keyspace="n", region="r"
        ) == admin_f.get_database(db_id_string, keyspace="n", region="r")

    @pytest.mark.describe(
        "test of token inheritance in spawning from AstraDBDatabaseAdmin"
    )
    def test_astradbdatabaseadmin_token_inheritance(self) -> None:
        opts_0 = defaultAPIOptions(environment=Environment.DEV)
        opts_t = opts_0.with_override(
            APIOptions(token=StaticTokenProvider("static")),
        )
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        opts_f = opts_0.with_override(APIOptions(token=token_f))
        adbadmin_t = AstraDBDatabaseAdmin(
            api_endpoint=api_ep0123_dev, api_options=opts_t
        )
        adbadmin_0 = AstraDBDatabaseAdmin(
            api_endpoint=api_ep0123_dev, api_options=opts_0
        )
        adbadmin_f = AstraDBDatabaseAdmin(
            api_endpoint=api_ep0123_dev, api_options=opts_f
        )

        assert adbadmin_t.get_database(
            token=token_f,
            keyspace="n",
        ) == adbadmin_f.get_database(keyspace="n")
        assert adbadmin_0.get_database(
            token=token_f,
            keyspace="n",
        ) == adbadmin_f.get_database(keyspace="n")

    @pytest.mark.describe(
        "test of token inheritance in spawning from DataAPIDatabaseAdmin"
    )
    def test_dataapidatabaseadmin_token_inheritance(self) -> None:
        a_e_string = "http://x.y:123"
        opts_0 = defaultAPIOptions(environment=Environment.PROD)
        opts_t = opts_0.with_override(
            APIOptions(token=StaticTokenProvider("static")),
        )
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        opts_f = opts_0.with_override(
            APIOptions(token=token_f),
        )
        dadbadmin_t = DataAPIDatabaseAdmin(api_endpoint=a_e_string, api_options=opts_t)
        dadbadmin_0 = DataAPIDatabaseAdmin(api_endpoint=a_e_string, api_options=opts_0)
        dadbadmin_f = DataAPIDatabaseAdmin(api_endpoint=a_e_string, api_options=opts_f)

        assert dadbadmin_t.get_database(
            token=token_f, keyspace="n"
        ) == dadbadmin_f.get_database(keyspace="n")
        assert dadbadmin_0.get_database(
            token=token_f, keyspace="n"
        ) == dadbadmin_f.get_database(keyspace="n")

    @pytest.mark.describe("test of token inheritance in spawning from Database, Astra")
    def test_database_token_inheritance_astra(self) -> None:
        opts_0 = defaultAPIOptions(environment=Environment.TEST)
        opts_t = opts_0.with_override(
            APIOptions(token=StaticTokenProvider("static")),
        )
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        opts_f = opts_0.with_override(
            APIOptions(token=token_f),
        )
        database_t = Database(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_t
        )
        a_database_t = AsyncDatabase(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_t
        )
        database_0 = Database(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_0
        )
        a_database_0 = AsyncDatabase(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_0
        )
        database_f = Database(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_f
        )
        a_database_f = AsyncDatabase(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_f
        )

        assert isinstance(a_database_0.get_database_admin(), AstraDBDatabaseAdmin)
        assert (
            database_t.get_database_admin(token=token_f)
            == database_f.get_database_admin()
        )
        assert (
            database_0.get_database_admin(token=token_f)
            == database_f.get_database_admin()
        )
        assert (
            a_database_t.get_database_admin(token=token_f)
            == a_database_f.get_database_admin()
        )
        assert (
            a_database_0.get_database_admin(token=token_f)
            == a_database_f.get_database_admin()
        )

    @pytest.mark.describe("test of option override in getting a DB admin from a DB")
    def test_database_get_database_admin_options(self) -> None:
        opts_0 = defaultAPIOptions(environment=Environment.TEST)
        database_0 = Database(
            api_endpoint=api_ep9999_test, keyspace="k", api_options=opts_0
        )

        db_admin_param = database_0.get_database_admin(token="tx")
        db_admin_opt = database_0.get_database_admin(
            spawn_api_options=APIOptions(token="tx"),
        )
        db_admin_opt_param = database_0.get_database_admin(
            token="tx",
            spawn_api_options=APIOptions(token="t_another"),
        )
        assert db_admin_param == db_admin_opt
        assert db_admin_param == db_admin_opt_param

        adatabase_0 = database_0.to_async()

        adb_admin_param = adatabase_0.get_database_admin(token="tx")
        adb_admin_opt = adatabase_0.get_database_admin(
            spawn_api_options=APIOptions(token="tx"),
        )
        adb_admin_opt_param = adatabase_0.get_database_admin(
            token="tx",
            spawn_api_options=APIOptions(token="t_another"),
        )
        assert adb_admin_param == adb_admin_opt
        assert adb_admin_param == adb_admin_opt_param

        assert adb_admin_param == db_admin_param

    @pytest.mark.describe(
        "test of token inheritance in spawning from Database, non-Astra"
    )
    def test_database_token_inheritance_nonastra(self) -> None:
        a_e_string = "http://x.y:123"
        opts_0 = defaultAPIOptions(environment=Environment.OTHER)
        opts_t = opts_0.with_override(
            APIOptions(token=StaticTokenProvider("static")),
        )
        token_f = UsernamePasswordTokenProvider(username="u", password="p")
        opts_f = opts_0.with_override(
            APIOptions(token=token_f),
        )
        database_t = Database(api_endpoint=a_e_string, keyspace="k", api_options=opts_t)
        a_database_t = AsyncDatabase(
            api_endpoint=a_e_string, keyspace="k", api_options=opts_t
        )
        database_0 = Database(api_endpoint=a_e_string, keyspace="k", api_options=opts_0)
        a_database_0 = AsyncDatabase(
            api_endpoint=a_e_string, keyspace="k", api_options=opts_0
        )
        database_f = Database(api_endpoint=a_e_string, keyspace="k", api_options=opts_f)
        a_database_f = AsyncDatabase(
            api_endpoint=a_e_string, keyspace="k", api_options=opts_f
        )

        assert isinstance(a_database_0.get_database_admin(), DataAPIDatabaseAdmin)
        assert (
            database_t.get_database_admin(token=token_f)
            == database_f.get_database_admin()
        )
        assert (
            database_0.get_database_admin(token=token_f)
            == database_f.get_database_admin()
        )
        assert (
            a_database_t.get_database_admin(token=token_f)
            == a_database_f.get_database_admin()
        )
        assert (
            a_database_0.get_database_admin(token=token_f)
            == a_database_f.get_database_admin()
        )

    @pytest.mark.describe(
        "test of id, endpoint, region normalization in get_database(_admin)"
    )
    def test_param_normalize_getdatabase(self) -> None:
        # the case of ID only is deferred to an integration test (it's impure)
        api_ep = "https://01234567-89ab-cdef-0123-456789abcdef-the-region.apps.astra.datastax.com"
        db_id = "01234567-89ab-cdef-0123-456789abcdef"
        db_reg = "the-region"
        opts = defaultAPIOptions(environment=Environment.PROD).with_override(
            APIOptions(token=StaticTokenProvider("t1")),
        )
        adm = AstraDBAdmin(api_options=opts)

        db_adm1 = adm.get_database_admin(db_id, region=db_reg)
        with pytest.raises(ValueError):
            adm.get_database_admin(api_ep, region=db_reg)
        db_adm3 = adm.get_database_admin(api_ep)
        with pytest.raises(ValueError):
            adm.get_database_admin(api_ep, region="not-that-one")

        assert db_adm1 == db_adm3

        db_1 = adm.get_database(db_id, region=db_reg, keyspace="the_ks")
        with pytest.raises(ValueError):
            adm.get_database(api_ep, region=db_reg, keyspace="the_ks")
        db_3 = adm.get_database(api_ep, keyspace="the_ks")
        with pytest.raises(ValueError):
            adm.get_database(api_ep, region="not-that-one", keyspace="the_ks")

        assert db_1 == db_3

    @pytest.mark.describe(
        "test of spawner_database for AstraDBDatabaseAdmin if not provided"
    )
    def test_spawnerdatabase_astradbdatabaseadmin_notprovided(self) -> None:
        api_ep = "https://01234567-89ab-cdef-0123-456789abcdef-the-region.apps.astra.datastax.com"
        db_adm = AstraDBDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.PROD),
        )
        assert db_adm.spawner_database.api_endpoint == api_ep

    @pytest.mark.describe(
        "test of spawner_database for DataAPIDatabaseAdmin if not provided"
    )
    def test_spawnerdatabase_dataapidatabaseadmin_notprovided(self) -> None:
        api_ep = "http://aa"
        db_adm = DataAPIDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.OTHER),
        )
        assert db_adm.spawner_database.api_endpoint == api_ep

    @pytest.mark.describe(
        "test of spawner_database for AstraDBDatabaseAdmin, sync db provided"
    )
    def test_spawnerdatabase_astradbdatabaseadmin_syncprovided(self) -> None:
        api_ep = "https://01234567-89ab-cdef-0123-456789abcdef-the-region.apps.astra.datastax.com"
        db = Database(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.PROD),
            keyspace="M",
        )
        db_adm = AstraDBDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.PROD),
            spawner_database=db,
        )
        assert db_adm.spawner_database is db

    @pytest.mark.describe(
        "test of spawner_database for AstraDBDatabaseAdmin, async db provided"
    )
    def test_spawnerdatabase_astradbdatabaseadmin_asyncprovided(self) -> None:
        api_ep = "https://01234567-89ab-cdef-0123-456789abcdef-the-region.apps.astra.datastax.com"
        adb = AsyncDatabase(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.PROD),
            keyspace="M",
        )
        db_adm = AstraDBDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.PROD),
            spawner_database=adb,
        )
        assert db_adm.spawner_database is adb

    @pytest.mark.describe(
        "test of spawner_database for DataAPIDatabaseAdmin, sync db provided"
    )
    def test_spawnerdatabase_dataapidatabaseadmin_syncprovided(self) -> None:
        api_ep = "http://aa"
        db = Database(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.OTHER),
            keyspace="k",
        )
        db_adm = DataAPIDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.OTHER),
            spawner_database=db,
        )
        assert db_adm.spawner_database is db

    @pytest.mark.describe(
        "test of spawner_database for DataAPIDatabaseAdmin, async db provided"
    )
    def test_spawnerdatabase_dataapidatabaseadmin_asyncprovided(self) -> None:
        api_ep = "http://aa"
        adb = AsyncDatabase(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.OTHER),
            keyspace="k",
        )
        db_adm = DataAPIDatabaseAdmin(
            api_endpoint=api_ep,
            api_options=defaultAPIOptions(environment=Environment.OTHER),
            spawner_database=adb,
        )
        assert db_adm.spawner_database is adb



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.constants import DefaultRowType
from astrapy.data.utils.table_converters import _TableConverterAgent
from astrapy.utils.api_commander import APICommander
from astrapy.utils.api_options import (
    SerdesOptions,
    defaultAPIOptions,
    defaultSerdesOptions,
)

from ..conftest import _repaint_NaNs
from ..table_decimal_support_assets import (
    BASELINE_COLUMNS,
    BASELINE_KEY_STR,
    BASELINE_OBJ,
    COLLTYPES_CUSTOM_OBJ_SCHEMA_TRIPLES,
    COLLTYPES_TRIPLE_IDS,
    WDECS_KEY_STR,
    WDECS_OBJ,
    WDECS_OBJ_COLUMNS,
)


class TestTableDecimalSupportUnit:
    @pytest.mark.describe("test of decimal-related conversions in table codec paths")
    def test_decimalsupport_table_codecpath(self) -> None:
        t_agent: _TableConverterAgent[DefaultRowType] = _TableConverterAgent(
            options=defaultAPIOptions(environment="prod").serdes_options,
        )
        # baseline, encode then decode and check
        baseline_fully_encoded = APICommander._decimal_aware_encode_payload(
            t_agent.preprocess_payload(BASELINE_OBJ, map2tuple_checker=None)
        )
        baseline_obj_2 = t_agent.postprocess_row(
            APICommander._decimal_aware_parse_json_response(baseline_fully_encoded),  # type: ignore[arg-type]
            columns_dict=BASELINE_COLUMNS,
            similarity_pseudocolumn=None,
        )
        assert _repaint_NaNs(baseline_obj_2) == _repaint_NaNs(BASELINE_OBJ)
        # with-decimals, encode then decode and check
        wdecs_fully_encoded = APICommander._decimal_aware_encode_payload(
            t_agent.preprocess_payload(WDECS_OBJ, map2tuple_checker=None)
        )
        wdecs_2 = t_agent.postprocess_row(
            APICommander._decimal_aware_parse_json_response(wdecs_fully_encoded),  # type: ignore[arg-type]
            columns_dict=WDECS_OBJ_COLUMNS,
            similarity_pseudocolumn=None,
        )
        assert _repaint_NaNs(wdecs_2) == _repaint_NaNs(WDECS_OBJ)

        # baseline, keys (decode only)
        baseline_kobj_2 = t_agent.postprocess_key(
            APICommander._decimal_aware_parse_json_response(BASELINE_KEY_STR),  # type: ignore[arg-type]
            primary_key_schema_dict=BASELINE_COLUMNS,
        )[1]
        assert _repaint_NaNs(baseline_kobj_2) == _repaint_NaNs(BASELINE_OBJ)
        # with-decimals, keys (decode only)
        wdecs_kobj_2 = t_agent.postprocess_key(
            APICommander._decimal_aware_parse_json_response(WDECS_KEY_STR),  # type: ignore[arg-type]
            primary_key_schema_dict=WDECS_OBJ_COLUMNS,
        )[1]
        assert _repaint_NaNs(wdecs_kobj_2) == _repaint_NaNs(WDECS_OBJ)

    @pytest.mark.parametrize(
        ("colltype_custom", "colltype_obj", "colltype_columns"),
        COLLTYPES_CUSTOM_OBJ_SCHEMA_TRIPLES,
        ids=COLLTYPES_TRIPLE_IDS,
    )
    @pytest.mark.describe(
        "test of decimal-related conversions in table codec paths, collection types"
    )
    def test_decimalsupport_table_collectiontypes_codecpath(
        self,
        colltype_custom: bool,
        colltype_obj: dict[str, Any],
        colltype_columns: dict[str, Any],
    ) -> None:
        t_agent: _TableConverterAgent[DefaultRowType] = _TableConverterAgent(
            options=defaultSerdesOptions.with_override(
                SerdesOptions(
                    custom_datatypes_in_reading=colltype_custom,
                    use_decimals_in_collections=True,
                ),
            ),
        )

        fully_encoded = APICommander._decimal_aware_encode_payload(
            t_agent.preprocess_payload(colltype_obj, map2tuple_checker=None)
        )
        colltype_obj_2 = t_agent.postprocess_row(
            APICommander._decimal_aware_parse_json_response(fully_encoded),  # type: ignore[arg-type]
            columns_dict=colltype_columns,
            similarity_pseudocolumn=None,
        )
        assert _repaint_NaNs(colltype_obj_2) == _repaint_NaNs(colltype_obj)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, Iterable

import pytest

from astrapy.data_types import DataAPIDictUDT


class TestDataAPIDictUDT:
    @pytest.mark.describe("test of DataAPIDictUDT instance checks")
    def test_dataapidictudt_instancechecks(self) -> None:
        test_dict = {"name": "John", "age": 40}
        test_udt = DataAPIDictUDT(test_dict)

        # equality
        assert test_dict == test_udt
        assert test_udt == test_dict

        # instance checks
        assert isinstance(test_udt, dict)
        assert isinstance(test_udt, DataAPIDictUDT)
        assert not isinstance(test_dict, DataAPIDictUDT)

        # construction methods
        test_udt_2 = DataAPIDictUDT(name="John", age=40)
        test_udt_3 = DataAPIDictUDT([("name", "John"), ("age", 40)])
        assert test_udt_2 == test_udt
        assert test_udt_3 == test_udt

        # dict behaviour
        def _t_eq(ite1: Iterable[Any], ite2: Iterable[Any]) -> bool:
            lst1 = list(ite1)
            lst2 = list(ite2)
            return len(lst1) == len(lst2) and all(itm1 in lst2 for itm1 in lst1)

        assert "name" in test_udt
        assert "xyz" not in test_udt
        assert _t_eq(test_udt.items(), test_dict.items())
        assert _t_eq(test_udt.keys(), test_dict.keys())
        assert _t_eq(test_udt.values(), test_dict.values())
        assert len(test_udt) == len(test_dict)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest
from pytest_httpserver import HTTPServer

from astrapy import Database, Table
from astrapy.exceptions import TooManyRowsToCountException
from astrapy.utils.api_options import defaultAPIOptions
from astrapy.utils.request_tools import HttpMethod

from ..conftest import DefaultAsyncTable, DefaultTable

BASE_PATH = "v1"
KEYSPACE = "keyspace"
TABLE_NAME = "table"
PATH_SUFFIX = f"{KEYSPACE}/{TABLE_NAME}"


@pytest.fixture
def mock_table(httpserver: HTTPServer) -> DefaultTable:
    base_endpoint = httpserver.url_for("/")
    table: DefaultTable = Table(
        database=Database(
            api_endpoint=base_endpoint,
            keyspace=KEYSPACE,
            api_options=defaultAPIOptions(environment="other"),
        ),
        name=TABLE_NAME,
        keyspace=None,
        api_options=defaultAPIOptions(environment="other"),
    )
    return table


@pytest.fixture
def mock_atable(httpserver: HTTPServer, mock_table: DefaultTable) -> DefaultAsyncTable:
    return mock_table.to_async()


class TestTableDryMethods:
    @pytest.mark.describe("test of table estimated_document_count and alias, sync")
    def test_table_estimated_document_count_sync(
        self,
        httpserver: HTTPServer,
        mock_table: DefaultTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        assert mock_table.estimated_document_count() == 500

    @pytest.mark.describe("test of table estimated_document_count and alias, async")
    async def test_table_estimated_document_count_async(
        self,
        httpserver: HTTPServer,
        mock_atable: DefaultAsyncTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        assert await mock_atable.estimated_document_count() == 500

    @pytest.mark.describe("test of table count_documents and alias, sync")
    def test_table_count_documents_sync(
        self,
        httpserver: HTTPServer,
        mock_table: DefaultTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        assert mock_table.count_documents({}, upper_bound=650) == 500

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 1000, "moreData": True}})
        with pytest.raises(TooManyRowsToCountException):
            mock_table.count_documents({}, upper_bound=1000)

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 50}})
        with pytest.raises(TooManyRowsToCountException):
            mock_table.count_documents({}, upper_bound=20)

    @pytest.mark.describe("test of table count_documents and alias, async")
    async def test_table_count_documents_async(
        self,
        httpserver: HTTPServer,
        mock_atable: DefaultAsyncTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        assert await mock_atable.count_documents({}, upper_bound=650) == 500

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 1000, "moreData": True}})
        with pytest.raises(TooManyRowsToCountException):
            await mock_atable.count_documents({}, upper_bound=1000)

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 50}})
        with pytest.raises(TooManyRowsToCountException):
            await mock_atable.count_documents({}, upper_bound=20)

    @pytest.mark.describe("test of table update_one, sync")
    def test_table_dry_update_one_sync(
        self,
        httpserver: HTTPServer,
        mock_table: DefaultTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {}})
        mock_table.update_one({"pk": "v"}, {"$set": {"x": {1, 2, 3}}})

    @pytest.mark.describe("test of table update_one, async")
    async def test_table_dry_update_one_async(
        self,
        httpserver: HTTPServer,
        mock_atable: DefaultAsyncTable,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {}})
        await mock_atable.update_one({"pk": "v"}, {"$set": {"x": {1, 2, 3}}})



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import AsyncDatabase
from astrapy.constants import ModelStatus
from astrapy.info import EmbeddingProvider, FindEmbeddingProvidersResult

from ..conftest import IS_ASTRA_DB, clean_nulls_from_dict


def _count_models(fep_result: FindEmbeddingProvidersResult) -> int:
    return len(
        [
            model
            for prov_v in fep_result.embedding_providers.values()
            for model in prov_v.models
        ]
    )


class TestVectorizeOpsAsync:
    @pytest.mark.describe("test of find_embedding_providers, async")
    async def test_findembeddingproviders_async(
        self,
        async_database: AsyncDatabase,
    ) -> None:
        database_admin = async_database.get_database_admin()
        ep_result = await database_admin.async_find_embedding_providers()

        assert isinstance(ep_result, FindEmbeddingProvidersResult)

        assert all(
            isinstance(emb_prov, EmbeddingProvider)
            for emb_prov in ep_result.embedding_providers.values()
        )

        reconstructed = {
            ep_name: EmbeddingProvider._from_dict(emb_prov.as_dict())
            for ep_name, emb_prov in ep_result.embedding_providers.items()
        }
        assert reconstructed == ep_result.embedding_providers
        dict_mapping = {
            ep_name: emb_prov.as_dict()
            for ep_name, emb_prov in ep_result.embedding_providers.items()
        }
        cleaned_dict_mapping = clean_nulls_from_dict(dict_mapping)
        cleaned_raw_info = clean_nulls_from_dict(
            ep_result.raw_info["embeddingProviders"]  # type: ignore[index]
        )
        if IS_ASTRA_DB:
            # TODO remove this (Astra-only) cleanup once apiModelSupport deployed
            raw_info_has_ams = any(
                "apiModelSupport" in model_dict
                for prov_v in cleaned_raw_info.values()
                for model_dict in prov_v["models"]
            )
            if not raw_info_has_ams:
                for emb_prov in cleaned_dict_mapping.values():
                    for model in emb_prov["models"]:
                        if "apiModelSupport" in model:
                            del model["apiModelSupport"]
        assert cleaned_dict_mapping == cleaned_raw_info

    # TODO: open to Astra DB once available
    @pytest.mark.skipif(IS_ASTRA_DB, reason="Filtering models not yet on Astra DB")
    @pytest.mark.describe("test of find_embedding_providers filtering, async")
    async def test_filtered_findembeddingproviders_async(
        self,
        async_database: AsyncDatabase,
    ) -> None:
        database_admin = async_database.get_database_admin()
        default_count = _count_models(
            await database_admin.async_find_embedding_providers()
        )

        all_count = _count_models(
            await database_admin.async_find_embedding_providers(filter_model_status="")
        )
        all_count_e = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status=ModelStatus.ALL,
            )
        )
        assert all_count_e == all_count

        sup_count = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status="SUPPORTED"
            )
        )
        dep_count = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status="DEPRECATED"
            )
        )
        eol_count = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status="END_OF_LIFE"
            )
        )
        sup_count_e = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status=ModelStatus.SUPPORTED,
            )
        )
        dep_count_e = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status=ModelStatus.DEPRECATED,
            )
        )
        eol_count_e = _count_models(
            await database_admin.async_find_embedding_providers(
                filter_model_status=ModelStatus.END_OF_LIFE,
            )
        )
        assert sup_count_e == sup_count
        assert dep_count_e == dep_count
        assert eol_count_e == eol_count

        assert sup_count + dep_count + eol_count == all_count
        assert sup_count == default_count



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from ..conftest import IS_ASTRA_DB, DefaultCollection

PJ = {"_id": True, "title": False}


class TestCollectionGeneralBM25Sync:
    @pytest.mark.skipif(IS_ASTRA_DB, reason="Feature not on Astra DB yet")
    @pytest.mark.describe("test of collection bm25 DML with vectorize, sync")
    def test_collection_bm25_dml_vectorize_sync(
        self,
        sync_empty_farr_vectorize_collection: DefaultCollection,
    ) -> None:
        coll = sync_empty_farr_vectorize_collection

        coll.insert_one(
            {
                "_id": "lx",
                "title": "Only lexical",
                "$lexical": "Water flows downward.",
            }
        )

        fol_doc = coll.find_one({"$lexical": {"$match": "water"}}, projection=PJ)
        assert fol_doc is not None
        assert fol_doc["_id"] == "lx"

        fcl_docs = coll.find({"$lexical": {"$match": "water"}}, projection=PJ).to_list()
        assert fcl_docs == [{"_id": "lx"}]

        # no lexical matching occurs
        fol_doc_n = coll.find_one({"$lexical": {"$match": "fire"}}, projection=PJ)
        assert fol_doc_n is None

        # cannot find a doc without vector
        fov_doc_n = coll.find_one(sort={"$vectorize": "water"}, projection=PJ)
        assert fov_doc_n is None

        coll.insert_many(
            [
                {
                    "_id": "hy",
                    "title": "Both, using hybrid",
                    "$hybrid": "Plants need water.",
                },
                {
                    "_id": "hy_0",
                    "title": "Both, using hybrid, irrelevant",
                    "$hybrid": "Sand forms dunes.",
                },
            ]
        )

        folv_doc = coll.find_one(
            {"$lexical": {"$match": "water"}},
            sort={"$vectorize": "water"},
            projection=PJ,
        )
        assert folv_doc is not None
        assert folv_doc["_id"] == "hy"

        # only one matches the lexical part and has a vector
        fcv_docs = coll.find(
            {"$lexical": {"$match": "water"}},
            sort={"$vectorize": "water"},
            projection=PJ,
        ).to_list()
        assert fcv_docs == [{"_id": "hy"}]

        coll.delete_many({})

        coll.insert_one(
            {
                "_id": "vl",
                "title": "Both, but different",
                "$lexical": "Birds fly.",
                "$vectorize": "Fish swim.",
            }
        )

        upd_one_result = coll.update_one(
            {"$lexical": {"$match": "birds"}},
            {"$set": {"updated": True}},
            sort={"$lexical": "fly"},
        )
        assert upd_one_result.update_info["n"] > 0

        upd_one_result_n = coll.update_one(
            {"$lexical": {"$match": "dogs"}},
            {"$set": {"updated": True}},
            sort={"$lexical": "fly"},
        )
        assert upd_one_result_n.update_info["n"] == 0

        upd_one_result_n2 = coll.update_one(
            {"$lexical": {"$match": "birds"}},
            {"$set": {"updated": True}},
            sort={"$lexical": "cats"},
        )
        assert upd_one_result_n2.update_info["n"] == 0

        f1u_result_doc = coll.find_one_and_update(
            {"$lexical": {"$match": "birds"}},
            {"$set": {"updated_n": 1}},
            sort={"$lexical": "fly"},
            projection=PJ,
        )
        assert f1u_result_doc is not None
        assert f1u_result_doc["_id"] == "vl"

        u1_result = coll.update_one(
            {"$lexical": {"$match": "birds"}},
            {"$set": {"updated_n": 2}},
            sort={"$lexical": "fly"},
        )
        assert u1_result.update_info["n"] > 0

        f1r_result_doc = coll.find_one_and_replace(
            {"$lexical": {"$match": "birds"}},
            {"name": "replaced 2"},
            sort={"$lexical": "fly"},
            projection=PJ,
        )
        assert f1r_result_doc is not None
        assert f1r_result_doc["_id"] == "vl"

        # Suspended until #2150 is understood/resolved
        # r1_result = coll.replace_one(
        #     {"$lexical": {"$match": "birds"}},
        #     {"name": "replaced 1"},
        #     sort={"$lexical": "fly"},
        # )
        # assert r1_result is not None
        # assert r1_result.update_info["n"] > 0

        coll.insert_many(
            [
                {
                    "_id": "vl2",
                    "title": "Again: both, but different",
                    "$lexical": "Ants tiptap.",
                    "$vectorize": "Snakes crawl.",
                },
                {
                    "_id": "vl3",
                    "title": "One more: Both, but different",
                    "$lexical": "Worms inch.",
                    "$vectorize": "Spiders climb.",
                },
            ]
        )

        f1d_result_doc = coll.find_one_and_delete(
            {"$lexical": {"$match": "ants"}},
            sort={"$lexical": "tiptap"},
            projection=PJ,
        )
        assert f1d_result_doc is not None
        assert f1d_result_doc["_id"] == "vl2"

        # Suspended until #2150 is understood/resolved
        # f1d_result_doc_n = coll.find_one_and_delete(
        #     {"$lexical": {"$match": "ants"}},
        #     sort={"$lexical": "tiptap"},
        #     projection=PJ,
        # )
        # assert f1d_result_doc is None

        d1_result = coll.delete_one(
            {"$lexical": {"$match": "worms"}},
            sort={"$lexical": "inch"},
        )
        assert d1_result.deleted_count > 0

        # Suspended until #2150 is understood/resolved
        # d1_result_n = coll.delete_one(
        #     {"$lexical": {"$match": "worms"}},
        #     sort={"$lexical": "inch"},
        # )
        # assert d1_result_n.deleted_count == 0

        coll.insert_many(
            [
                {
                    "_id": "hy2",
                    "title": "Document Alpha for *many test",
                    "$hybrid": "The King is dead.",
                },
                {
                    "_id": "hy3",
                    "title": "Document Omega for *many test",
                    "$hybrid": "Long live the king!",
                },
                {
                    "_id": "hy4_n",
                    "title": "Document Negative for *many test",
                    "$hybrid": "Vive la Republique!",
                },
            ]
        )

        um_result = coll.update_many(
            {"$lexical": {"$match": "king"}},
            {"$set": {"updated": True}},
        )
        assert um_result.update_info["n"] == 2
        assert coll.count_documents({"updated": True}, upper_bound=5) == 2

        dm_result = coll.delete_many(
            {"$lexical": {"$match": "king"}},
        )
        assert dm_result.deleted_count == 2
        assert (
            coll.count_documents(
                {"$lexical": {"$match": "king"}},
                upper_bound=5,
            )
            == 0
        )

        dm_result_n = coll.delete_many(
            {"$lexical": {"$match": "king"}},
        )
        assert dm_result_n.deleted_count == 0
        assert (
            coll.count_documents(
                {"$lexical": {"$match": "king"}},
                upper_bound=5,
            )
            == 0
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import math
import os
from decimal import Decimal
from typing import Any, Dict, Iterable

import pytest

from astrapy import (
    AsyncCollection,
    AsyncTable,
    Collection,
    Database,
    Table,
)
from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import VectorMetric
from astrapy.data_types import DataAPIMap, DataAPISet
from astrapy.info import (
    CollectionDefinition,
    CollectionLexicalOptions,
    CollectionRerankOptions,
    CollectionVectorOptions,
    RerankServiceOptions,
)
from astrapy.utils.unset import _UNSET, UnsetType

from ..conftest import (
    ADMIN_ENV_LIST,
    ADMIN_ENV_VARIABLE_MAP,
    CQL_AVAILABLE,
    HEADER_EMBEDDING_API_KEY_OPENAI,
    HEADER_RERANKING_API_KEY_NVIDIA,
    IS_ASTRA_DB,
    SECONDARY_KEYSPACE,
    DataAPICredentials,
    DataAPICredentialsInfo,
    async_fail_if_not_removed,
    clean_nulls_from_dict,
    is_future_version,
    sync_fail_if_not_removed,
)
from .table_structure_assets import (
    TEST_ALL_RETURNS_TABLE_DEFINITION,
    TEST_ALL_RETURNS_TABLE_NAME,
    TEST_ALLMAPS_TABLE_DEFINITION,
    TEST_ALLMAPS_TABLE_NAME,
    TEST_COLLINDEXED_TABLE_DEFINITION,
    TEST_COLLINDEXED_TABLE_INDEXES,
    TEST_COLLINDEXED_TABLE_NAME,
    TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_COLUMN,
    TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_NAME,
    TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_OPTIONS,
    TEST_COMPOSITE_TABLE_DEFINITION,
    TEST_COMPOSITE_TABLE_NAME,
    TEST_COMPOSITE_TABLE_VECTOR_INDEX_COLUMN,
    TEST_COMPOSITE_TABLE_VECTOR_INDEX_NAME,
    TEST_COMPOSITE_TABLE_VECTOR_INDEX_OPTIONS,
    TEST_KMS_VECTORIZE_TABLE_DEFINITION,
    TEST_KMS_VECTORIZE_TABLE_NAME,
    TEST_KMS_VECTORIZE_TABLE_VECTOR_INDEX_NAME,
    TEST_LOGICALFILTERING_TABLE_DEFINITION,
    TEST_LOGICALFILTERING_TABLE_INDEXES,
    TEST_LOGICALFILTERING_TABLE_NAME,
    TEST_MULTIPLEVECTORIZE_TABLE_DEFINITION,
    TEST_MULTIPLEVECTORIZE_TABLE_INDEX_OPTIONS,
    TEST_MULTIPLEVECTORIZE_TABLE_INDEXES,
    TEST_MULTIPLEVECTORIZE_TABLE_NAME,
    TEST_SIMPLE_TABLE_DEFINITION,
    TEST_SIMPLE_TABLE_NAME,
    TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN,
    TEST_SIMPLE_TABLE_VECTOR_INDEX_NAME,
    TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS,
    TEST_SIMPLE_UDT_DEFINITION,
    TEST_SIMPLE_UDT_NAME,
    TEST_TEXTINDEX_INDEX_COLUMN,
    TEST_TEXTINDEX_INDEX_NAME,
    TEST_TEXTINDEX_TABLE_DEFINITION,
    TEST_TEXTINDEX_TABLE_NAME,
    TEST_VECTORIZE_TABLE_DEFINITION,
    TEST_VECTORIZE_TABLE_NAME,
    TEST_VECTORIZE_TABLE_VECTOR_INDEX_NAME,
    VECTORIZE_TEXTS,
)
from .table_udt_assets import (
    EXTENDED_PLAYER_TABLE_DEFINITION,
    EXTENDED_PLAYER_TABLE_NAME,
    EXTENDED_PLAYER_TYPE_DEFINITION,
    EXTENDED_PLAYER_TYPE_NAME,
    PLAYER_TABLE_DEFINITION,
    PLAYER_TABLE_NAME,
    PLAYER_TYPE_DEFINITION,
    PLAYER_TYPE_NAME,
    TEST_UDTCOLLINDEXED_TABLE_DEFINITION,
    TEST_UDTCOLLINDEXED_TABLE_INDEXES,
    TEST_UDTCOLLINDEXED_TABLE_NAME,
)

DefaultCollection = Collection[Dict[str, Any]]
DefaultAsyncCollection = AsyncCollection[Dict[str, Any]]
DefaultTable = Table[Dict[str, Any]]
DefaultAsyncTable = AsyncTable[Dict[str, Any]]

TEST_COLLECTION_INSTANCE_NAME = "test_coll_instance"
TEST_COLLECTION_NAME = "id_test_collection"
TEST_SERVICE_COLLECTION_NAME = "test_indepth_vectorize_collection"
TEST_TABLE_INSTANCE_NAME = "test_tbl_instance"
TEST_FARR_VECTORIZE_COLLECTION_NAME = "test_farr_collection_vectorize"
TEST_FARR_VECTOR_COLLECTION_NAME = "test_farr_collection_vector"

_NaN = object()
_DNaN = object()


def _repaint_NaNs(val: Any) -> Any:
    if isinstance(val, float) and math.isnan(val):
        return _NaN
    if isinstance(val, Decimal) and math.isnan(val):
        return _DNaN
    elif isinstance(val, dict):
        return {_repaint_NaNs(k): _repaint_NaNs(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [_repaint_NaNs(x) for x in val]
    elif isinstance(val, DataAPISet):
        return DataAPISet(_repaint_NaNs(v) for v in val)
    elif isinstance(val, DataAPIMap):
        return DataAPIMap((_repaint_NaNs(k), _repaint_NaNs(v)) for k, v in val.items())
    elif isinstance(val, set):
        return {_repaint_NaNs(v) for v in val}
    elif isinstance(val, dict):
        return {_repaint_NaNs(k): _repaint_NaNs(v) for k, v in val.items()}
    else:
        return val


def _typify_tuple(tpl: tuple[Any, ...]) -> tuple[Any, ...]:
    return tuple([(v, type(v)) for v in tpl])


@pytest.fixture(scope="session")
def sync_collection_instance(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultCollection]:
    """Just an instance of the class, no DB-level stuff."""
    yield sync_database.get_collection(TEST_COLLECTION_INSTANCE_NAME)


@pytest.fixture(scope="function")
def async_collection_instance(
    sync_collection_instance: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """Just an instance of the class, no DB-level stuff."""
    yield sync_collection_instance.to_async()


@pytest.fixture(scope="session")
def sync_collection(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultCollection]:
    """An actual collection on DB, in the main keyspace"""
    collection = sync_database.create_collection(
        TEST_COLLECTION_NAME,
        definition=CollectionDefinition(
            indexing={"deny": ["not_indexed"]},
            vector=CollectionVectorOptions(
                dimension=2,
                metric=VectorMetric.COSINE,
            ),
        ),
        spawn_api_options=APIOptions(
            serdes_options=SerdesOptions(
                binary_encode_vectors=True,
                custom_datatypes_in_reading=True,
                unroll_iterables_to_lists=True,
            ),
        ),
    )
    yield collection

    sync_database.drop_collection(TEST_COLLECTION_NAME)


@pytest.fixture(scope="function")
def sync_empty_collection(
    sync_collection: DefaultCollection,
) -> Iterable[DefaultCollection]:
    """Emptied for each test function"""
    sync_collection.delete_many({})
    yield sync_collection


@pytest.fixture(scope="function")
def async_collection(
    sync_collection: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """An actual collection on DB, the same as the sync counterpart"""
    yield sync_collection.to_async()


@pytest.fixture(scope="function")
def async_empty_collection(
    sync_empty_collection: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """Emptied for each test function"""
    yield sync_empty_collection.to_async()


@pytest.fixture(scope="session")
def service_collection_parameters() -> Iterable[dict[str, Any]]:
    yield {
        "dimension": 1536,
        "provider": "openai",
        "modelName": "text-embedding-ada-002",
        "api_key": HEADER_EMBEDDING_API_KEY_OPENAI,
        "reranking_api_key": HEADER_RERANKING_API_KEY_NVIDIA,
    }


@pytest.fixture(scope="session")
def sync_service_collection(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
    service_collection_parameters: dict[str, Any],
) -> Iterable[DefaultCollection]:
    """
    An actual collection on DB, in the main keyspace.
    """
    params = service_collection_parameters
    collection = sync_database.create_collection(
        TEST_SERVICE_COLLECTION_NAME,
        definition=(
            CollectionDefinition.builder()
            .set_vector_metric(VectorMetric.DOT_PRODUCT)
            .set_vector_service(
                provider=params["provider"],
                model_name=params["modelName"],
            )
            .build()
        ),
        embedding_api_key=params["api_key"],
    )
    yield collection

    sync_database.drop_collection(TEST_SERVICE_COLLECTION_NAME)


@pytest.fixture(scope="function")
def sync_empty_service_collection(
    sync_service_collection: DefaultCollection,
) -> Iterable[DefaultCollection]:
    """Emptied for each test function"""
    sync_service_collection.delete_many({})
    yield sync_service_collection


@pytest.fixture(scope="function")
def async_empty_service_collection(
    sync_empty_service_collection: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """Emptied for each test function"""
    yield sync_empty_service_collection.to_async()


@pytest.fixture(scope="session")
def rerankservice_collection_parameters() -> Iterable[CollectionRerankOptions]:
    yield CollectionRerankOptions(
        service=RerankServiceOptions(
            provider="nvidia",
            model_name="nvidia/llama-3.2-nv-rerankqa-1b-v2",
        ),
    )


@pytest.fixture(scope="session")
def lexical_collection_parameters() -> Iterable[CollectionLexicalOptions]:
    yield CollectionLexicalOptions(
        analyzer="STANDARD",
    )


@pytest.fixture(scope="module")
def sync_farr_vectorize_collection(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
    service_collection_parameters: dict[str, Any],
    rerankservice_collection_parameters: CollectionRerankOptions,
    lexical_collection_parameters: CollectionLexicalOptions,
) -> Iterable[DefaultCollection]:
    """
    An actual collection on DB, in the main keyspace.
    """
    params = service_collection_parameters

    # reranker credentials through header is needed for non-Astra (HCD) runs
    reranking_api_key: str | UnsetType
    if "ASTRAPY_FINDANDRERANK_USE_RERANKER_HEADER" in os.environ:
        assert params["reranking_api_key"] is not None
        reranking_api_key = params["reranking_api_key"]
    else:
        reranking_api_key = _UNSET

    collection = sync_database.create_collection(
        TEST_FARR_VECTORIZE_COLLECTION_NAME,
        definition=(
            CollectionDefinition.builder()
            .set_vector_metric(VectorMetric.DOT_PRODUCT)
            .set_vector_service(
                provider=params["provider"],
                model_name=params["modelName"],
            )
            .set_rerank(rerankservice_collection_parameters)
            .set_lexical(lexical_collection_parameters)
            .build()
        ),
        embedding_api_key=params["api_key"],
        reranking_api_key=reranking_api_key,
    )
    yield collection

    sync_database.drop_collection(TEST_FARR_VECTORIZE_COLLECTION_NAME)


@pytest.fixture(scope="function")
def sync_empty_farr_vectorize_collection(
    sync_farr_vectorize_collection: DefaultCollection,
) -> Iterable[DefaultCollection]:
    """Emptied for each test function"""
    sync_farr_vectorize_collection.delete_many({})
    yield sync_farr_vectorize_collection


@pytest.fixture(scope="function")
def async_empty_farr_vectorize_collection(
    sync_empty_farr_vectorize_collection: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """Emptied for each test function"""
    yield sync_empty_farr_vectorize_collection.to_async()


@pytest.fixture(scope="module")
def sync_farr_vector_collection(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
    service_collection_parameters: dict[str, Any],
    rerankservice_collection_parameters: CollectionRerankOptions,
    lexical_collection_parameters: CollectionLexicalOptions,
) -> Iterable[DefaultCollection]:
    """
    An actual collection on DB, in the main keyspace.
    """
    params = service_collection_parameters

    # reranker credentials through header is needed for non-Astra (HCD) runs
    reranking_api_key: str | UnsetType
    if "ASTRAPY_FINDANDRERANK_USE_RERANKER_HEADER" in os.environ:
        assert params["reranking_api_key"] is not None
        reranking_api_key = params["reranking_api_key"]
    else:
        reranking_api_key = _UNSET

    collection = sync_database.create_collection(
        TEST_FARR_VECTOR_COLLECTION_NAME,
        definition=(
            CollectionDefinition.builder()
            .set_vector_metric(VectorMetric.DOT_PRODUCT)
            .set_vector_dimension(2)
            .set_rerank(rerankservice_collection_parameters)
            .set_lexical(lexical_collection_parameters)
            .build()
        ),
        reranking_api_key=reranking_api_key,
    )
    yield collection

    sync_database.drop_collection(TEST_FARR_VECTOR_COLLECTION_NAME)


@pytest.fixture(scope="function")
def sync_empty_farr_vector_collection(
    sync_farr_vector_collection: DefaultCollection,
) -> Iterable[DefaultCollection]:
    """Emptied for each test function"""
    sync_farr_vector_collection.delete_many({})
    yield sync_farr_vector_collection


@pytest.fixture(scope="function")
def async_empty_farr_vector_collection(
    sync_empty_farr_vector_collection: DefaultCollection,
) -> Iterable[DefaultAsyncCollection]:
    """Emptied for each test function"""
    yield sync_empty_farr_vector_collection.to_async()


@pytest.fixture(scope="session")
def simple_udt(
    sync_database: Database,
) -> Iterable[str]:
    """Creation (and then cleanup) of a UDT. Returns the type name."""
    sync_database.create_type(
        TEST_SIMPLE_UDT_NAME,
        definition=TEST_SIMPLE_UDT_DEFINITION,
    )
    yield TEST_SIMPLE_UDT_NAME

    sync_database.drop_type(TEST_SIMPLE_UDT_NAME)


@pytest.fixture(scope="session")
def player_udt(
    sync_database: Database,
) -> Iterable[str]:
    """Creation (and then cleanup) of a UDT. Returns the type name."""
    sync_database.create_type(
        PLAYER_TYPE_NAME,
        definition=PLAYER_TYPE_DEFINITION,
    )
    yield PLAYER_TYPE_NAME

    sync_database.drop_type(PLAYER_TYPE_NAME)


@pytest.fixture(scope="session")
def extended_player_udt(
    sync_database: Database,
) -> Iterable[str]:
    """Creation (and then cleanup) of a UDT. Returns the type name."""
    sync_database.create_type(
        EXTENDED_PLAYER_TYPE_NAME,
        definition=EXTENDED_PLAYER_TYPE_DEFINITION,
    )
    yield EXTENDED_PLAYER_TYPE_NAME

    sync_database.drop_type(EXTENDED_PLAYER_TYPE_NAME)


@pytest.fixture(scope="session")
def sync_table_instance(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """Just an instance of the class, no DB-level stuff."""
    yield sync_database.get_table(TEST_TABLE_INSTANCE_NAME)


@pytest.fixture(scope="function")
def async_table_instance(
    sync_table_instance: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Just an instance of the class, no DB-level stuff."""
    yield sync_table_instance.to_async()


@pytest.fixture(scope="session")
def sync_table_all_returns(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_ALL_RETURNS_TABLE_NAME,
        definition=TEST_ALL_RETURNS_TABLE_DEFINITION,
    )
    yield table

    sync_database.drop_table(TEST_ALL_RETURNS_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_all_returns(
    sync_table_all_returns: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_all_returns.delete_many({})
    yield sync_table_all_returns


@pytest.fixture(scope="function")
def async_table_all_returns(
    sync_table_all_returns: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_all_returns.to_async()


@pytest.fixture(scope="function")
def async_empty_table_all_returns(
    sync_empty_table_all_returns: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_all_returns.to_async()


@pytest.fixture(scope="session")
def sync_table_simple(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_SIMPLE_TABLE_NAME,
        definition=TEST_SIMPLE_TABLE_DEFINITION,
    )
    table.create_vector_index(
        TEST_SIMPLE_TABLE_VECTOR_INDEX_NAME,
        TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN,
        options=TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS,
    )
    yield table

    sync_database.drop_table(TEST_SIMPLE_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_simple(
    sync_table_simple: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_simple.delete_many({})
    yield sync_table_simple


@pytest.fixture(scope="function")
def async_table_simple(
    sync_table_simple: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_simple.to_async()


@pytest.fixture(scope="function")
def async_empty_table_simple(
    sync_empty_table_simple: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_simple.to_async()


@pytest.fixture(scope="session")
def sync_table_composite(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_COMPOSITE_TABLE_NAME,
        definition=TEST_COMPOSITE_TABLE_DEFINITION,
    )
    table.create_vector_index(
        TEST_COMPOSITE_TABLE_VECTOR_INDEX_NAME,
        TEST_COMPOSITE_TABLE_VECTOR_INDEX_COLUMN,
        options=TEST_COMPOSITE_TABLE_VECTOR_INDEX_OPTIONS,
    )
    table.create_index(
        TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_NAME,
        TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_COLUMN,
        options=TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_OPTIONS,
    )
    yield table

    sync_database.drop_table(TEST_COMPOSITE_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_composite(
    sync_table_composite: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_composite.delete_many({})
    yield sync_table_composite


@pytest.fixture(scope="function")
def async_table_composite(
    sync_table_composite: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_composite.to_async()


@pytest.fixture(scope="function")
def async_empty_table_composite(
    sync_empty_table_composite: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_composite.to_async()


@pytest.fixture(scope="session")
def sync_table_vectorize(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_VECTORIZE_TABLE_NAME,
        definition=TEST_VECTORIZE_TABLE_DEFINITION,
    )
    table.create_vector_index(
        TEST_VECTORIZE_TABLE_VECTOR_INDEX_NAME,
        TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN,
        options=TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS,
    )
    yield table

    sync_database.drop_table(TEST_VECTORIZE_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_vectorize(
    sync_table_vectorize: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_vectorize.delete_many({})
    yield sync_table_vectorize


@pytest.fixture(scope="function")
def async_table_vectorize(
    sync_table_vectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_vectorize.to_async()


@pytest.fixture(scope="function")
def async_empty_table_vectorize(
    sync_empty_table_vectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_vectorize.to_async()


###
@pytest.fixture(scope="session")
def sync_table_multiplevectorize(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
    service_collection_parameters: dict[str, Any],
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    params = service_collection_parameters

    table = sync_database.create_table(
        TEST_MULTIPLEVECTORIZE_TABLE_NAME,
        definition=TEST_MULTIPLEVECTORIZE_TABLE_DEFINITION,
        embedding_api_key=params["api_key"],
    )
    for index_name, index_column in TEST_MULTIPLEVECTORIZE_TABLE_INDEXES:
        table.create_vector_index(
            index_name,
            index_column,
            options=TEST_MULTIPLEVECTORIZE_TABLE_INDEX_OPTIONS,
        )

    yield table

    sync_database.drop_table(TEST_MULTIPLEVECTORIZE_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_multiplevectorize(
    sync_table_multiplevectorize: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_multiplevectorize.delete_many({})
    yield sync_table_multiplevectorize


@pytest.fixture(scope="function")
def async_table_multiplevectorize(
    sync_table_multiplevectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_multiplevectorize.to_async()


@pytest.fixture(scope="function")
def async_empty_table_multiplevectorize(
    sync_empty_table_multiplevectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_multiplevectorize.to_async()


###


@pytest.fixture(scope="session")
def sync_table_kms_vectorize(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_KMS_VECTORIZE_TABLE_NAME,
        definition=TEST_KMS_VECTORIZE_TABLE_DEFINITION,
    )
    table.create_vector_index(
        TEST_KMS_VECTORIZE_TABLE_VECTOR_INDEX_NAME,
        TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN,
        options=TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS,
    )
    yield table

    sync_database.drop_table(TEST_KMS_VECTORIZE_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_kms_vectorize(
    sync_table_kms_vectorize: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_kms_vectorize.delete_many({})
    yield sync_table_kms_vectorize


@pytest.fixture(scope="function")
def async_table_kms_vectorize(
    sync_table_kms_vectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_kms_vectorize.to_async()


@pytest.fixture(scope="function")
def async_empty_table_kms_vectorize(
    sync_empty_table_kms_vectorize: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_kms_vectorize.to_async()


@pytest.fixture(scope="session")
def sync_table_allmaps(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_ALLMAPS_TABLE_NAME,
        definition=TEST_ALLMAPS_TABLE_DEFINITION,
    )
    yield table

    sync_database.drop_table(TEST_ALLMAPS_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_allmaps(
    sync_table_allmaps: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_allmaps.delete_many({})
    yield sync_table_allmaps


@pytest.fixture(scope="function")
def async_table_allmaps(
    sync_table_allmaps: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_allmaps.to_async()


@pytest.fixture(scope="function")
def async_empty_table_allmaps(
    sync_empty_table_allmaps: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_allmaps.to_async()


@pytest.fixture(scope="session")
def sync_table_udt_player(
    sync_database: Database,
    player_udt: str,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        PLAYER_TABLE_NAME,
        definition=PLAYER_TABLE_DEFINITION,
    )
    yield table

    sync_database.drop_table(PLAYER_TABLE_NAME)


@pytest.fixture(scope="session")
def sync_table_udt_extended_player(
    sync_database: Database,
    extended_player_udt: str,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        EXTENDED_PLAYER_TABLE_NAME,
        definition=EXTENDED_PLAYER_TABLE_DEFINITION,
    )
    yield table

    sync_database.drop_table(EXTENDED_PLAYER_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_udt_player(
    sync_table_udt_player: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_udt_player.delete_many({})
    yield sync_table_udt_player


@pytest.fixture(scope="function")
def sync_empty_table_udt_extended_player(
    sync_table_udt_extended_player: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_udt_extended_player.delete_many({})
    yield sync_table_udt_extended_player


@pytest.fixture(scope="function")
def async_empty_table_udt_player(
    sync_empty_table_udt_player: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_udt_player.to_async()


@pytest.fixture(scope="function")
def async_empty_table_udt_extended_player(
    sync_empty_table_udt_extended_player: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_udt_extended_player.to_async()


@pytest.fixture(scope="session")
def sync_table_udtcollindexed(
    sync_database: Database,
    player_udt: str,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_UDTCOLLINDEXED_TABLE_NAME,
        definition=TEST_UDTCOLLINDEXED_TABLE_DEFINITION,
    )
    for index_name, index_column in TEST_UDTCOLLINDEXED_TABLE_INDEXES:
        table.create_index(index_name, index_column)

    yield table

    sync_database.drop_table(TEST_UDTCOLLINDEXED_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_udtcollindexed(
    sync_table_udtcollindexed: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_udtcollindexed.delete_many({})
    yield sync_table_udtcollindexed


@pytest.fixture(scope="function")
def async_empty_table_udtcollindexed(
    sync_empty_table_udtcollindexed: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_udtcollindexed.to_async()


@pytest.fixture(scope="session")
def sync_table_collindexed(
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_COLLINDEXED_TABLE_NAME,
        definition=TEST_COLLINDEXED_TABLE_DEFINITION,
    )
    for index_name, index_column in TEST_COLLINDEXED_TABLE_INDEXES:
        table.create_index(index_name, index_column)

    yield table

    sync_database.drop_table(TEST_COLLINDEXED_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_collindexed(
    sync_table_collindexed: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_collindexed.delete_many({})
    yield sync_table_collindexed


@pytest.fixture(scope="function")
def async_empty_table_collindexed(
    sync_empty_table_collindexed: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_collindexed.to_async()


@pytest.fixture(scope="session")
def sync_table_textindex(
    data_api_credentials_kwargs: DataAPICredentials,
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_TEXTINDEX_TABLE_NAME,
        definition=TEST_TEXTINDEX_TABLE_DEFINITION,
    )
    table.create_text_index(
        TEST_TEXTINDEX_INDEX_NAME,
        TEST_TEXTINDEX_INDEX_COLUMN,
    )
    yield table

    sync_database.drop_table(TEST_TEXTINDEX_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_textindex(
    sync_table_textindex: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_textindex.delete_many({})
    yield sync_table_textindex


@pytest.fixture(scope="function")
def async_table_textindex(
    sync_table_textindex: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """An actual table on DB, the same as the sync counterpart"""
    yield sync_table_textindex.to_async()


@pytest.fixture(scope="function")
def async_empty_table_textindex(
    sync_empty_table_textindex: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_textindex.to_async()


@pytest.fixture(scope="session")
def sync_table_logicalfiltering(
    sync_database: Database,
) -> Iterable[DefaultTable]:
    """An actual table on DB, in the main keyspace"""
    table = sync_database.create_table(
        TEST_LOGICALFILTERING_TABLE_NAME,
        definition=TEST_LOGICALFILTERING_TABLE_DEFINITION,
    )
    for index_name, index_column in TEST_LOGICALFILTERING_TABLE_INDEXES:
        table.create_index(index_name, index_column)

    yield table

    sync_database.drop_table(TEST_LOGICALFILTERING_TABLE_NAME)


@pytest.fixture(scope="function")
def sync_empty_table_logicalfiltering(
    sync_table_logicalfiltering: DefaultTable,
) -> Iterable[DefaultTable]:
    """Emptied for each test function"""
    sync_table_logicalfiltering.delete_many({})
    yield sync_table_logicalfiltering


@pytest.fixture(scope="function")
def async_empty_table_logicalfiltering(
    sync_empty_table_logicalfiltering: DefaultTable,
) -> Iterable[DefaultAsyncTable]:
    """Emptied for each test function"""
    yield sync_empty_table_logicalfiltering.to_async()


__all__ = [
    "DataAPICredentials",
    "DataAPICredentialsInfo",
    "async_fail_if_not_removed",
    "clean_nulls_from_dict",
    "is_future_version",
    "sync_fail_if_not_removed",
    "HEADER_EMBEDDING_API_KEY_OPENAI",
    "IS_ASTRA_DB",
    "ADMIN_ENV_LIST",
    "ADMIN_ENV_VARIABLE_MAP",
    "CQL_AVAILABLE",
    "SECONDARY_KEYSPACE",
    "VECTORIZE_TEXTS",
    "_repaint_NaNs",
    "_typify_tuple",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pickle

import pytest

from astrapy.data_types import DataAPISet


class TestDataAPISet:
    @pytest.mark.describe("test of table set usage with hashables")
    def test_dataapiset_hashables(self) -> None:
        ts0: DataAPISet[int] = DataAPISet()
        assert ts0 == DataAPISet()
        assert set(ts0) == set()
        # identity/equality
        ts1 = DataAPISet([1, 2, 3])
        assert ts1 == DataAPISet([1, 2, 3])
        assert ts1 == DataAPISet([1, 2, 3, 2])
        assert ts1 != DataAPISet([1, 3, 2])
        assert set(ts1) == {1, 2, 3}
        # set operations
        assert ts1 - {2} == DataAPISet([1, 3])
        assert ts1 | {2} == DataAPISet([1, 2, 3])
        assert ts1 | {4} == DataAPISet([1, 2, 3, 4])
        assert ts1 | DataAPISet([4, 2, 5]) == DataAPISet([1, 2, 3, 4, 5])
        assert ts1 ^ DataAPISet([4, 2, 5]) == DataAPISet([1, 3, 4, 5])
        assert ts1 & DataAPISet([4, 2, 5]) == DataAPISet([2])

    @pytest.mark.describe("test of table set usage with non-hashables")
    def test_dataapiset_nonhashables(self) -> None:
        ts0: DataAPISet[list[int]] = DataAPISet()
        assert ts0 == DataAPISet()
        # identity/equality
        ts1 = DataAPISet([[1], [2], [3]])
        assert ts1 == DataAPISet([[1], [2], [3]])
        assert ts1 == DataAPISet([[1], [2], [3], [2]])
        assert ts1 != DataAPISet([[1], [3], [2]])
        # set operations
        assert ts1 - DataAPISet([[2]]) == DataAPISet([[1], [3]])
        assert ts1 | DataAPISet([[2]]) == DataAPISet([[1], [2], [3]])
        assert ts1 | DataAPISet([[4]]) == DataAPISet([[1], [2], [3], [4]])
        assert ts1 | DataAPISet([[4], [2], [5]]) == DataAPISet(
            [[1], [2], [3], [4], [5]]
        )
        assert ts1 ^ DataAPISet([[4], [2], [5]]) == DataAPISet([[1], [3], [4], [5]])
        assert ts1 & DataAPISet([[4], [2], [5]]) == DataAPISet([[2]])

    @pytest.mark.describe("test pickling of DataAPISet")
    def test_dataapiset_pickle(self) -> None:
        the_set = DataAPISet([1, 2, 3])
        assert pickle.loads(pickle.dumps(the_set)) == the_set



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from astrapy.data.info.table_descriptor.table_columns import (
    TableColumnTypeDescriptor,
    TableScalarColumnTypeDescriptor,
)
from astrapy.data.utils.table_types import ColumnType
from astrapy.utils.parsing import _warn_residual_keys


@dataclass
class CreateTypeDefinition:
    """
    A structure expressing the definition of a user-defined type( UDT) to be created
    through the Data API. This object is passed as the `definition` parameter to
    the database `create_type` method.

    See the Data API specifications for detailed specification and allowed values.

    Instances of this object can be created in three ways: using a fluent interface,
    passing a fully-formed definition to the class constructor, or coercing an
    appropriately-shaped plain dictionary into this class.

    Attributes:
        fields: a map from field names to their type definition object. This follows
            the same structure as the `columns` attribute of `CreateTableDefinition`.

    Example:
            >>> from astrapy.info import CreateTypeDefinition
            >>> from astrapy.info import ColumnType, TableScalarColumnTypeDescriptor
            >>>
            >>> type_definition_0 = (
            ...     CreateTypeDefinition.builder()
            ...     .add_field("tagline", ColumnType.TEXT)
            ...     .add_field("score", ColumnType.INT)
            ...     .add_field("height", "float")  # plain strings accepted for field types
            ...     .build()
            ... )
            >>>
            >>> type_definition_1 = CreateTypeDefinition(fields={
            ...     "tagline": TableScalarColumnTypeDescriptor(ColumnType.TEXT),
            ...     "score": TableScalarColumnTypeDescriptor(ColumnType.INT),
            ...     "height": TableScalarColumnTypeDescriptor(ColumnType.FLOAT),
            ... })
            >>>
            >>> fields_dict_2 = {
            ...     "fields": {
            ...         "tagline": "text",
            ...         "score": "int",
            ...         "height": "float",
            ...     },
            ... }
            >>> type_definition_2 = CreateTypeDefinition.coerce(fields_dict_2)
            >>>
            >>> fields_dict_3 = {
            ...     "fields": {
            ...         "tagline": "text",
            ...         "score": {"type": "int"},
            ...         "height": TableScalarColumnTypeDescriptor(ColumnType.FLOAT),
            ...     },
            ... }
            >>> type_definition_3_mixed = CreateTypeDefinition.coerce(fields_dict_3)
            >>> type_definition_0 == type_definition_1
            True
            >>> type_definition_1 == type_definition_2
            True
            >>> type_definition_2 == type_definition_3_mixed
            True
    """

    fields: dict[str, TableColumnTypeDescriptor]

    def __init__(
        self,
        *,
        fields: dict[str, TableColumnTypeDescriptor | dict[str, Any] | str],
    ) -> None:
        self.fields = {
            fld_n: TableColumnTypeDescriptor.coerce(fld_v)
            for fld_n, fld_v in fields.items()
        }

    def __repr__(self) -> str:
        fld_desc = f"fields=[{','.join(sorted(self.fields.keys()))}]"
        return f"{self.__class__.__name__}({fld_desc})"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            "fields": {col_n: col_v.as_dict() for col_n, col_v in self.fields.items()},
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> CreateTypeDefinition:
        """
        Create an instance of CreateTypeDefinition from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(cls, raw_dict, {"fields"})
        return CreateTypeDefinition(
            fields={
                fld_n: TableColumnTypeDescriptor.coerce(fld_v)
                for fld_n, fld_v in raw_dict["fields"].items()
            },
        )

    @classmethod
    def coerce(
        cls, raw_input: CreateTypeDefinition | dict[str, Any]
    ) -> CreateTypeDefinition:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into a CreateTypeDefinition.
        """

        if isinstance(raw_input, CreateTypeDefinition):
            return raw_input
        else:
            return cls._from_dict(raw_input)

    @staticmethod
    def builder() -> CreateTypeDefinition:
        """
        Create an "empty" builder for constructing a type definition through
        a fluent interface. The resulting object has no fields yet: those are to
        be added progressively with the `add_field` method.

        Being a "type without fields", the type definition returned by this method
        cannot be directly used to create a type.

        See the class docstring for a full example on using the fluent interface.

        Returns:
            a CreateTypeDefinition formally describing a type without fields.
        """

        return CreateTypeDefinition(fields={})

    def add_field(
        self, field_name: str, field_type: str | ColumnType
    ) -> CreateTypeDefinition:
        """
        Return a new type definition object with an added field
        of a scalar type (i.e. not a list, set or other composite type).
        This method is for use within the fluent interface for progressively
        building the final type definition.

        See the class docstring for a full example on using the fluent interface.

        Args:
            field_name: the name of the new field to add to the type.
            field_type: a string, or a `ColumnType` value, defining
                the scalar type for the field.

        Returns:
            a CreateTypeDefinition obtained by adding (or replacing) the desired
            field to this type definition.
        """

        return CreateTypeDefinition(
            fields={
                **self.fields,
                **{
                    field_name: TableScalarColumnTypeDescriptor(
                        column_type=ColumnType.coerce(field_type)
                    )
                },
            },
        )

    def build(self) -> CreateTypeDefinition:
        """
        The final step in the fluent (builder) interface. Calling this method
        finalizes the definition that has been built so far and makes it into a
        type definition ready for use e.g. with the database's `create_type` method.

        See the class docstring for a full example on using the fluent interface.

        Returns:
            a CreateTypeDefinition obtained by finalizing the definition being
                built so far.
        """

        return self



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time

import pytest

from astrapy import Database
from astrapy.admin.admin import fetch_database_info
from astrapy.exceptions import DataAPITimeoutException, DevOpsAPITimeoutException

from ..conftest import IS_ASTRA_DB, DefaultCollection


class TestCollectionTimeoutSync:
    @pytest.mark.describe("test of collection count_documents timeout, sync")
    def test_collection_count_documents_timeout_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        sync_empty_collection.insert_many([{"a": 1}] * 500)
        time.sleep(2)

        with pytest.raises(DataAPITimeoutException) as exc:
            sync_empty_collection.count_documents(
                {}, upper_bound=800, general_method_timeout_ms=1
            )
        assert sync_empty_collection.count_documents({}, upper_bound=800) >= 500
        assert exc.value.timeout_type in {"connect", "read"}
        assert exc.value.endpoint is not None
        assert exc.value.raw_payload is not None

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.describe("test of database info timeout, sync")
    def test_database_info_timeout_sync(
        self,
        sync_database: Database,
    ) -> None:
        info = fetch_database_info(
            sync_database.api_endpoint,
            token=sync_database.api_options.token,
            keyspace=sync_database.keyspace,
        )
        assert info is not None

        with pytest.raises(DevOpsAPITimeoutException) as exc:
            info = fetch_database_info(
                sync_database.api_endpoint,
                token=sync_database.api_options.token,
                keyspace=sync_database.keyspace,
                request_timeout_ms=1,
            )
            assert info is not None
        assert exc.value.timeout_type in {"connect", "read"}
        assert exc.value.endpoint is not None
        assert exc.value.raw_payload is not None

    @pytest.mark.describe("test of cursor-based overall timeouts, sync")
    def test_cursor_overalltimeout_exceptions_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection
        col.insert_many([{"a": 1}] * 1000)

        col.distinct("a", general_method_timeout_ms=20000)
        with pytest.raises(DataAPITimeoutException):
            col.distinct("a", timeout_ms=1)

        col.distinct("a", general_method_timeout_ms=20000)
        with pytest.raises(DataAPITimeoutException):
            col.distinct("a", general_method_timeout_ms=1)

    @pytest.mark.describe("test of insert_many timeouts, sync")
    def test_insert_many_timeout_exceptions_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        fifty_docs = [{"seq": i} for i in range(50)]
        sync_collection.insert_many(
            fifty_docs, ordered=True, general_method_timeout_ms=20000
        )
        sync_collection.insert_many(
            fifty_docs, ordered=False, concurrency=1, general_method_timeout_ms=20000
        )
        sync_collection.insert_many(
            fifty_docs, ordered=False, concurrency=2, general_method_timeout_ms=20000
        )

        with pytest.raises(DataAPITimeoutException):
            sync_collection.insert_many(
                fifty_docs, ordered=True, general_method_timeout_ms=2
            )
        with pytest.raises(DataAPITimeoutException):
            sync_collection.insert_many(
                fifty_docs, ordered=False, concurrency=1, general_method_timeout_ms=2
            )
        with pytest.raises(DataAPITimeoutException):
            sync_collection.insert_many(
                fifty_docs, ordered=False, concurrency=2, general_method_timeout_ms=2
            )

    @pytest.mark.describe("test of update_many timeouts, sync")
    def test_update_many_timeout_exceptions_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        fifty_docs = [{"seq": i, "f": "update_many"} for i in range(50)]
        sync_collection.insert_many(fifty_docs, ordered=False, concurrency=3)

        sync_collection.update_many({"f": "update_many"}, {"$inc": {"seq": 100}})
        sync_collection.update_many(
            {"f": "update_many"},
            {"$inc": {"seq": 100}},
            general_method_timeout_ms=20000,
        )

        with pytest.raises(DataAPITimeoutException):
            sync_collection.update_many(
                {"f": "update_many"},
                {"$inc": {"seq": 100}},
                general_method_timeout_ms=2,
            )

    @pytest.mark.describe("test of delete_many timeouts, sync")
    def test_delete_many_timeout_exceptions_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        fifty_docs1 = [{"seq": i, "f": "delete_many1"} for i in range(50)]
        fifty_docs2 = [{"seq": i, "f": "delete_many2"} for i in range(50)]
        fifty_docs3 = [{"seq": i, "f": "delete_many3"} for i in range(50)]
        sync_collection.insert_many(
            fifty_docs1 + fifty_docs2 + fifty_docs3,
            ordered=False,
            concurrency=5,
        )

        sync_collection.delete_many({"f": "delete_many1"})
        sync_collection.delete_many(
            {"f": "delete_many2"}, general_method_timeout_ms=20000
        )
        with pytest.raises(DataAPITimeoutException):
            sync_collection.delete_many(
                {"f": "delete_many3"}, general_method_timeout_ms=2
            )

    @pytest.mark.describe("test of collection find-with-collective timeout, sync")
    def test_collection_find_with_collective_timeout_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        sync_empty_collection.insert_many([{"a": 1}] * 55)
        time.sleep(1)

        with pytest.raises(DataAPITimeoutException):
            sync_empty_collection.distinct("a", general_method_timeout_ms=1)

        cur_tl = sync_empty_collection.find()
        with pytest.raises(DataAPITimeoutException):
            cur_tl.to_list(general_method_timeout_ms=1)

        cur_fe = sync_empty_collection.find()
        with pytest.raises(DataAPITimeoutException):
            cur_fe.for_each(lambda doc: None, general_method_timeout_ms=1)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import AsyncDatabase, Database

from ..conftest import (
    IS_ASTRA_DB,
)


@pytest.mark.skipif(IS_ASTRA_DB, reason="Not supported on Astra DB")
class TestNonAstraAdmin:
    @pytest.mark.describe(
        "test of the keyspace crud with non-Astra DataAPIDatabaseAdmin, sync"
    )
    def test_nonastra_database_admin_keyspaces_sync(
        self, sync_database: Database
    ) -> None:
        """
        Test plan
        - database -> get_database_admin
            - list keyspaces, check name not there
            - create a keyspace
            - list keyspaces, check
            - database -> create_collection/list_collection_names
            - drop keyspace
            - list keyspaces, check
            - ALSO: test the update_db_keyspace flag when creating from db->dbadmin
        This version uses v2.0-compliant 'keyspace' method names
        """
        NEW_KS_NAME = "tnaa_test_ks_s"
        NEW_COLL_NAME = "tnaa_test_coll"
        database_admin = sync_database.get_database_admin()

        keyspaces1 = set(database_admin.list_keyspaces())
        assert NEW_KS_NAME not in keyspaces1

        database_admin.create_keyspace(NEW_KS_NAME)

        keyspaces2 = set(database_admin.list_keyspaces())
        assert NEW_KS_NAME in keyspaces2

        sync_database_ks = sync_database.with_options(keyspace=NEW_KS_NAME)
        sync_database_ks.create_collection(NEW_COLL_NAME)
        assert NEW_COLL_NAME in sync_database_ks.list_collection_names()

        database_admin.drop_keyspace(NEW_KS_NAME)

        keyspaces3 = set(database_admin.list_keyspaces())
        assert keyspaces3 == keyspaces1

        # update_db_keyspace:
        NEW_KS_NAME_NOT_UPDATED = "tnudn_notupd"
        NEW_KS_NAME_UPDATED = "tnudn_upd"

        database = sync_database._copy()

        keyspace0 = database.keyspace
        database_admin = database.get_database_admin()
        database_admin.create_keyspace(NEW_KS_NAME_NOT_UPDATED)
        assert database.keyspace == keyspace0

        database_admin.create_keyspace(NEW_KS_NAME_UPDATED, update_db_keyspace=True)
        assert database.keyspace == NEW_KS_NAME_UPDATED

        database_admin.drop_keyspace(NEW_KS_NAME_NOT_UPDATED)
        database_admin.drop_keyspace(NEW_KS_NAME_UPDATED)

    @pytest.mark.describe(
        "test of the keyspace crud with non-Astra DataAPIDatabaseAdmin, async"
    )
    async def test_nonastra_database_admin_keyspaces_async(
        self, async_database: AsyncDatabase
    ) -> None:
        """
        Test plan
        - database -> get_database_admin
            - list keyspaces, check name not there
            - create a keyspace
            - list keyspaces, check
            - database -> create_collection/list_collection_names
            - drop keyspace
            - list keyspaces, check
            - ALSO: test the update_db_keyspace flag when creating from db->dbadmin
        This version uses v2.0-compliant 'keyspace' method names
        """
        NEW_KS_NAME = "tnaa_test_ks_a"
        NEW_COLL_NAME = "tnaa_test_coll"
        database_admin = async_database.get_database_admin()

        keyspaces1 = set(await database_admin.async_list_keyspaces())
        assert NEW_KS_NAME not in keyspaces1

        await database_admin.async_create_keyspace(NEW_KS_NAME)

        keyspaces2 = set(await database_admin.async_list_keyspaces())
        assert NEW_KS_NAME in keyspaces2

        async_database_ks = async_database.with_options(keyspace=NEW_KS_NAME)
        await async_database_ks.create_collection(NEW_COLL_NAME)
        assert NEW_COLL_NAME in (await async_database_ks.list_collection_names())

        await database_admin.async_drop_keyspace(NEW_KS_NAME)

        keyspaces3 = set(await database_admin.async_list_keyspaces())
        assert keyspaces3 == keyspaces1

        # update_db_keyspace:
        NEW_KS_NAME_NOT_UPDATED = "tnudn_notupd"
        NEW_KS_NAME_UPDATED = "tnudn_upd"

        adatabase = async_database._copy()

        keyspace0 = adatabase.keyspace
        database_admin = adatabase.get_database_admin()
        await database_admin.async_create_keyspace(NEW_KS_NAME_NOT_UPDATED)
        assert adatabase.keyspace == keyspace0

        await database_admin.async_create_keyspace(
            NEW_KS_NAME_UPDATED, update_db_keyspace=True
        )
        assert adatabase.keyspace == NEW_KS_NAME_UPDATED

        await database_admin.async_drop_keyspace(NEW_KS_NAME_NOT_UPDATED)
        await database_admin.async_drop_keyspace(NEW_KS_NAME_UPDATED)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import pickle

import pytest

from astrapy.data_types import DataAPIDuration

# constants lifted from https://github.com/apache/cassandra/blob/
#   198df38860d1575b60fbcaf9a8a506d2efd1bb31/test/unit/org/apache/
#   cassandra/cql3/DurationTest.java#L37
NANOS_PER_MICRO = 1000
NANOS_PER_MILLI = 1000 * NANOS_PER_MICRO
NANOS_PER_SECOND = 1000 * NANOS_PER_MILLI
NANOS_PER_MINUTE = 60 * NANOS_PER_SECOND
NANOS_PER_HOUR = 60 * NANOS_PER_MINUTE


class TestDataAPIDuration:
    @pytest.mark.describe("test of duration type, errors in parsing from string")
    def test_dataapiduration_parse_errors(self) -> None:
        # baselines
        DataAPIDuration.from_string("P1Y")
        DataAPIDuration.from_string("P3W")
        # spurious begin
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("?P1Y")
        # spurious end
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1YR")
        # spurious in the middle
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1YG1M")
        # fake unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1B")
        # repeated unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y2Y1M")
        # out-of-order unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1M4Y")
        # quantity with '+'
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y+1M")
        # signs in the middle
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y-1M")
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("-P1Y-1M")

        # W-format with other things pre-T
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P3W1D")
        # W-format with other things post-T
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P3WT12H")
        # W-format with trailing T
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P3WT")

        # repeated unit, subday part
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1MT1H1M4M1.123S")
        # out-of-order unit, subday part
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1MT1H1M1.123S3H")
        # quantity with '+', subday part
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1MT1H1M+1.123S")
        # signs in the middle, subday part
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("P1Y1MT1H-1M1.123S")
        with pytest.raises(ValueError):
            DataAPIDuration.from_string("-P1Y1MT1H-1M1.123S")

    @pytest.mark.describe("test of duration type, null-duration forms parsing")
    def test_dataapiduration_parse_nulldurations(self) -> None:
        zerod = DataAPIDuration(
            signum=+1,
            months=0,
            days=0,
            nanoseconds=0,
        )
        assert zerod == DataAPIDuration.from_string("P")
        assert zerod == DataAPIDuration.from_string("PT")
        assert zerod == DataAPIDuration.from_string("-P")
        assert zerod == DataAPIDuration.from_string("-PT")

    @pytest.mark.describe("test of duration type, lifecycle")
    def test_dataapiduration_lifecycle(self) -> None:
        # base
        dfull = DataAPIDuration.from_string("P1Y1M1DT1H1M1.001001001S")
        dfull_exp = DataAPIDuration(
            signum=+1, months=13, days=1, nanoseconds=3661001001001
        )
        assert dfull == dfull_exp
        # holey cases
        dhole1 = DataAPIDuration.from_string("P1M1DT1M")
        dhole1_exp = DataAPIDuration(
            signum=+1, months=1, days=1, nanoseconds=60000000000
        )
        assert dhole1 == dhole1_exp
        dhole2 = DataAPIDuration.from_string("-P1YT1H1.000000123S")
        dhole2_exp = DataAPIDuration(
            signum=-1, months=12, days=0, nanoseconds=3601000000123
        )
        assert dhole2 == dhole2_exp
        assert DataAPIDuration.from_string("PT1M1S") == DataAPIDuration.from_string(
            "PT61S"
        )
        # equivalent-formulation identity
        deq1 = DataAPIDuration.from_string("P12MT60M")
        deq2 = DataAPIDuration.from_string("P1YT1H")
        assert deq1 == deq2
        # null duration
        zd0 = DataAPIDuration(signum=+1, months=0, days=0, nanoseconds=0)
        zd1 = DataAPIDuration.from_string("P0Y")
        zd2 = DataAPIDuration.from_string("P0M0D")
        zd3 = DataAPIDuration.from_string("-P0DT0S")
        assert zd0 == zd1
        assert zd0 == zd2
        assert zd0 == zd3
        # stringy functions
        dfull.to_string()
        repr(dfull)
        zd0.to_string()
        repr(zd0)
        # trailing 'T' test
        one_y_0 = DataAPIDuration.from_string("P1YT")
        one_y_t = DataAPIDuration.from_string("P1Y")
        m_one_y_0 = DataAPIDuration.from_string("-P1YT")
        m_one_y_t = DataAPIDuration.from_string("-P1Y")
        assert one_y_0 == one_y_t
        assert m_one_y_0 == m_one_y_t

    @pytest.mark.describe("test of duration type, errors in c-parsing from string")
    def test_dataapiduration_c_parse_errors(self) -> None:
        # baseline
        DataAPIDuration.from_c_string("1mo")
        # spurious begin
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("X1mo")
        # spurious end
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1moY")
        # spurious in the middle
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1moX1s")
        # fake unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1mo6b1s")
        # repeated unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1mo1d1d1s")
        # out-of-order unit
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1d1mo1s")
        # quantity with '+'
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1mo+3s")
        # signs in the middle
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("-1d-1s")
        with pytest.raises(ValueError):
            DataAPIDuration.from_c_string("1d-1s")

    @pytest.mark.describe("test of duration type, c-lifecycle")
    def test_dataapiduration_c_lifecycle(self) -> None:
        # base
        dfull = DataAPIDuration.from_c_string("1y1mo1w1d1h1m1s1ms1us1ns")
        dfull_exp = DataAPIDuration(
            signum=+1, months=13, days=8, nanoseconds=3661001001001
        )
        assert dfull == dfull_exp
        # mu, signed quantity and case-insensitivity
        assert DataAPIDuration.from_c_string("-123US") == DataAPIDuration.from_c_string(
            "-123µs"
        )
        # holey cases
        dhole1 = DataAPIDuration.from_c_string("1mo1d1m1ms1ns")
        dhole1_exp = DataAPIDuration(
            signum=+1, months=1, days=1, nanoseconds=60001000001
        )
        assert dhole1 == dhole1_exp
        dhole2 = DataAPIDuration.from_c_string("-1y1w1h1s1us")
        dhole2_exp = DataAPIDuration(
            signum=-1, months=12, days=7, nanoseconds=3601000001000
        )
        assert dhole2 == dhole2_exp
        assert DataAPIDuration.from_c_string("1m1s") == DataAPIDuration.from_c_string(
            "61s"
        )
        # equivalent-formulation identity
        deq1 = DataAPIDuration.from_c_string("13mo2w1h1s1us")
        deq2 = DataAPIDuration.from_c_string("1y1mo14d60m1000ms1000ns")
        assert deq1 == deq2
        # null duration
        zd0 = DataAPIDuration(signum=+1, months=0, days=0, nanoseconds=0)
        zd1 = DataAPIDuration.from_c_string("0y")
        zd2 = DataAPIDuration.from_c_string("0mo0d")
        zd3 = DataAPIDuration.from_c_string("0w0ns")
        zd4 = DataAPIDuration.from_c_string("")
        assert zd0 == zd1
        assert zd0 == zd2
        assert zd0 == zd3
        assert zd0 == zd4
        # stringy functions
        dfull.to_c_string()
        repr(dfull)
        zd0.to_c_string()
        repr(zd0)

    @pytest.mark.describe("test of duration type, c-parsing and parsing equivalence")
    def test_dataapiduration_c_parsing_parsing_equivalence(self) -> None:
        c_td_ok_0 = DataAPIDuration.from_c_string("")
        c_td_ok_1 = DataAPIDuration.from_c_string("-1h1s333ms")
        c_td_ok_2 = DataAPIDuration.from_c_string("-191h1s")
        c_td_ok_3 = DataAPIDuration.from_c_string("1h44m3s777000ns")
        c_td_no = DataAPIDuration.from_c_string("-1y1s")

        td_ok_0 = DataAPIDuration.from_string("PT0S")
        td_ok_1 = DataAPIDuration.from_string("-PT1H1.333S")
        td_ok_2 = DataAPIDuration.from_string("-PT191H1S")
        td_ok_3 = DataAPIDuration.from_string("PT1H44M3.000777S")
        td_no = DataAPIDuration.from_string("-P1YT1S")

        assert c_td_ok_0 == td_ok_0
        assert c_td_ok_1 == td_ok_1
        assert c_td_ok_2 == td_ok_2
        assert c_td_ok_3 == td_ok_3
        assert c_td_no == td_no

    @pytest.mark.describe("test of duration type, timedelta conversions")
    def test_dataapiduration_timedelta_conversions(self) -> None:
        td_ok_0 = DataAPIDuration.from_string("PT0S")
        td_ok_1 = DataAPIDuration.from_string("-PT1H1.333S")
        td_ok_2 = DataAPIDuration.from_string("-PT191H1S")
        td_ok_3 = DataAPIDuration.from_string("PT1H44M3.000777S")
        td_no = DataAPIDuration.from_string("-P1YT1S")

        # due to lossy conversions and the month-day-subday math being different,
        # other (more challenging) values will fail here (limitations of timedelta)
        assert DataAPIDuration.from_timedelta(td_ok_0.to_timedelta()) == td_ok_0
        assert DataAPIDuration.from_timedelta(td_ok_1.to_timedelta()) == td_ok_1
        assert DataAPIDuration.from_timedelta(td_ok_2.to_timedelta()) == td_ok_2
        assert DataAPIDuration.from_timedelta(td_ok_3.to_timedelta()) == td_ok_3
        with pytest.raises(ValueError):
            td_no.to_timedelta()

        tdelta_0 = datetime.timedelta()
        tdelta_1 = datetime.timedelta(days=1, seconds=2, microseconds=345678)
        tdelta_2 = datetime.timedelta(seconds=10, milliseconds=-456)
        assert DataAPIDuration.from_timedelta(tdelta_0).to_timedelta() == tdelta_0
        assert DataAPIDuration.from_timedelta(tdelta_1).to_timedelta() == tdelta_1
        assert DataAPIDuration.from_timedelta(tdelta_2).to_timedelta() == tdelta_2

    @pytest.mark.describe("test of duration parsing, lifted from Cassandra codebase")
    def test_dataapiduration_parsing_cassandra_testcases(self) -> None:
        assert DataAPIDuration(+1, 12, 2, 0) == DataAPIDuration.from_string("P1Y2D")
        assert DataAPIDuration(+1, 14, 0, 0) == DataAPIDuration.from_string("P1Y2M")
        assert DataAPIDuration(+1, 0, 14, 0) == DataAPIDuration.from_string("P2W")
        assert DataAPIDuration(
            +1, 12, 0, 2 * NANOS_PER_HOUR
        ) == DataAPIDuration.from_string("P1YT2H")
        assert DataAPIDuration(-1, 14, 0, 0) == DataAPIDuration.from_string("-P1Y2M")
        assert DataAPIDuration(+1, 0, 2, 0) == DataAPIDuration.from_string("P2D")
        assert DataAPIDuration(
            +1, 0, 0, 30 * NANOS_PER_HOUR
        ) == DataAPIDuration.from_string("PT30H")
        assert DataAPIDuration(
            +1, 0, 0, 30 * NANOS_PER_HOUR + 20 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_string("PT30H20M")
        assert DataAPIDuration(
            +1, 0, 0, 20 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_string("PT20M")
        assert DataAPIDuration(
            +1, 0, 0, 56 * NANOS_PER_SECOND
        ) == DataAPIDuration.from_string("PT56S")
        assert DataAPIDuration(
            +1, 15, 0, 130 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_string("P1Y3MT2H10M")

    @pytest.mark.describe("test of duration c-parsing, lifted from Cassandra codebase")
    def test_dataapiduration_c_parsing_cassandra_testcases(self) -> None:
        assert DataAPIDuration(+1, 14, 0, 0) == DataAPIDuration.from_c_string("1y2mo")
        assert DataAPIDuration(-1, 14, 0, 0) == DataAPIDuration.from_c_string("-1y2mo")
        assert DataAPIDuration(+1, 14, 0, 0) == DataAPIDuration.from_c_string("1Y2MO")
        assert DataAPIDuration(+1, 0, 14, 0) == DataAPIDuration.from_c_string("2w")
        assert DataAPIDuration(
            +1, 0, 2, 10 * NANOS_PER_HOUR
        ) == DataAPIDuration.from_c_string("2d10h")
        assert DataAPIDuration(+1, 0, 2, 0) == DataAPIDuration.from_c_string("2d")
        assert DataAPIDuration(
            +1, 0, 0, 30 * NANOS_PER_HOUR
        ) == DataAPIDuration.from_c_string("30h")
        assert DataAPIDuration(
            +1, 0, 0, 30 * NANOS_PER_HOUR + 20 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_c_string("30h20m")
        assert DataAPIDuration(
            +1, 0, 0, 20 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_c_string("20m")
        assert DataAPIDuration(
            +1, 0, 0, 56 * NANOS_PER_SECOND
        ) == DataAPIDuration.from_c_string("56s")
        assert DataAPIDuration(
            +1, 0, 0, 567 * NANOS_PER_MILLI
        ) == DataAPIDuration.from_c_string("567ms")
        assert DataAPIDuration(
            +1, 0, 0, 1950 * NANOS_PER_MICRO
        ) == DataAPIDuration.from_c_string("1950us")
        assert DataAPIDuration(
            +1, 0, 0, 1950 * NANOS_PER_MICRO
        ) == DataAPIDuration.from_c_string("1950µs")
        assert DataAPIDuration(+1, 0, 0, 1950000) == DataAPIDuration.from_c_string(
            "1950000ns"
        )
        assert DataAPIDuration(+1, 0, 0, 1950000) == DataAPIDuration.from_c_string(
            "1950000NS"
        )
        assert DataAPIDuration(-1, 0, 0, 1950000) == DataAPIDuration.from_c_string(
            "-1950000ns"
        )
        assert DataAPIDuration(
            +1, 15, 0, 130 * NANOS_PER_MINUTE
        ) == DataAPIDuration.from_c_string("1y3mo2h10m")

    @pytest.mark.describe("test pickling of DataAPIDuration")
    def test_dataapiduration_pickle(self) -> None:
        the_duration = DataAPIDuration(signum=-1, months=1, days=2, nanoseconds=456)
        assert pickle.loads(pickle.dumps(the_duration)) == the_duration



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.api_options import APIOptions, SerdesOptions

from ..conftest import DefaultTable, dict_equal_same_class
from .table_row_assets import (
    ALLMAPS_CUSTOMTYPES_EMPTY_ROW,
    ALLMAPS_CUSTOMTYPES_ROW,
    ALLMAPS_STDLIB_EMPTY_ROW,
    ALLMAPS_STDLIB_ROW,
    DISTINCT_AR_ROWS,
    DISTINCT_EMPTYMAP_AR_ROWS,
)

MAP2TUPLES_OPTIONS = APIOptions(
    serdes_options=SerdesOptions(encode_maps_as_lists_in_tables="ALWAYS")
)
STDLIB_OPTIONS = APIOptions(
    serdes_options=SerdesOptions(custom_datatypes_in_reading=False)
)


class TestTableMapsAsTuplesSync:
    @pytest.mark.describe("test of table maps-as-tuples, insert one and find one, sync")
    def test_table_mapsastuples_insert_one_find_one_sync(
        self,
        sync_empty_table_allmaps: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_allmaps.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        customdt_table = tuplified_table
        stdlib_table = tuplified_table.with_options(api_options=STDLIB_OPTIONS)

        # custom types and corresponding serdes options
        cd_io_res = customdt_table.insert_one(ALLMAPS_CUSTOMTYPES_ROW)
        assert cd_io_res.inserted_id == {"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}
        assert (
            customdt_table.find_one({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
            == ALLMAPS_CUSTOMTYPES_ROW
        )

        # stdlib types and corresponding serdes options
        sl_io_res = stdlib_table.insert_one(ALLMAPS_STDLIB_ROW)
        assert sl_io_res.inserted_id == {"id": ALLMAPS_STDLIB_ROW["id"]}
        assert (
            stdlib_table.find_one({"id": ALLMAPS_STDLIB_ROW["id"]})
            == ALLMAPS_STDLIB_ROW
        )

    @pytest.mark.describe("test of table maps-as-tuples, empty-map insertions, sync")
    def test_table_mapsastuples_emptymaps_insertions_sync(
        self,
        sync_empty_table_allmaps: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_allmaps.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        customdt_table = tuplified_table
        stdlib_table = tuplified_table.with_options(api_options=STDLIB_OPTIONS)

        tuplified_table.insert_many(
            [
                ALLMAPS_CUSTOMTYPES_EMPTY_ROW,
                ALLMAPS_STDLIB_EMPTY_ROW,
            ]
        )
        tuplified_table.insert_one(ALLMAPS_CUSTOMTYPES_EMPTY_ROW)
        tuplified_table.insert_one(ALLMAPS_STDLIB_EMPTY_ROW)

        ct_e_ctrow = customdt_table.find_one(
            {"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        st_e_ctrow = customdt_table.find_one(
            {"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        dict_equal_same_class(ct_e_ctrow, st_e_ctrow)

        ct_e_strow = stdlib_table.find_one(
            {"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        st_e_strow = stdlib_table.find_one(
            {"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        dict_equal_same_class(ct_e_strow, st_e_strow)

    @pytest.mark.describe("test of table maps-as-tuples, insert many and find, sync")
    def test_table_mapsastuples_insert_many_find_sync(
        self,
        sync_empty_table_allmaps: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_allmaps.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        customdt_table = tuplified_table
        stdlib_table = tuplified_table.with_options(api_options=STDLIB_OPTIONS)

        # custom types and corresponding serdes options I: ordered insert
        cd_im_res = customdt_table.insert_many([ALLMAPS_CUSTOMTYPES_ROW], ordered=True)
        assert cd_im_res.inserted_ids == [{"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}]
        customdt_founds = customdt_table.find(
            {"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}
        ).to_list()
        assert customdt_founds == [ALLMAPS_CUSTOMTYPES_ROW]

        # custom types and corresponding serdes options II: unordered, concurrency=1 insert
        customdt_table.delete_many({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
        cd_im_res = customdt_table.insert_many(
            [ALLMAPS_CUSTOMTYPES_ROW], ordered=False, concurrency=1
        )
        assert cd_im_res.inserted_ids == [{"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}]
        customdt_founds = customdt_table.find(
            {"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}
        ).to_list()
        assert customdt_founds == [ALLMAPS_CUSTOMTYPES_ROW]

        # custom types and corresponding serdes options III: unordered, concurrency>1 insert
        customdt_table.delete_many({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
        cd_im_res = customdt_table.insert_many(
            [ALLMAPS_CUSTOMTYPES_ROW], ordered=False, concurrency=2
        )
        assert cd_im_res.inserted_ids == [{"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}]
        customdt_founds = customdt_table.find(
            {"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}
        ).to_list()
        assert customdt_founds == [ALLMAPS_CUSTOMTYPES_ROW]

        # stdlib types and corresponding serdes options
        sl_im_res = stdlib_table.insert_many([ALLMAPS_STDLIB_ROW])
        assert sl_im_res.inserted_ids == [{"id": ALLMAPS_STDLIB_ROW["id"]}]
        stdlib_founds = stdlib_table.find({"id": ALLMAPS_STDLIB_ROW["id"]}).to_list()
        assert stdlib_founds == [ALLMAPS_STDLIB_ROW]

    @pytest.mark.describe(
        "test of table maps-as-tuples, update one and delete one, sync"
    )
    def test_table_mapsastuples_update_one_delete_one_sync(
        self,
        sync_empty_table_allmaps: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_allmaps.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        customdt_table = tuplified_table
        stdlib_table = tuplified_table.with_options(api_options=STDLIB_OPTIONS)

        # custom types and corresponding serdes options
        customdt_table.insert_one({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
        customdt_table.update_one(
            {"id": ALLMAPS_CUSTOMTYPES_ROW["id"]},
            update={
                "$set": {k: v for k, v in ALLMAPS_CUSTOMTYPES_ROW.items() if k != "id"}
            },
        )
        assert (
            customdt_table.find_one({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
            == ALLMAPS_CUSTOMTYPES_ROW
        )
        customdt_table.delete_one({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]})
        assert customdt_table.find_one({"id": ALLMAPS_CUSTOMTYPES_ROW["id"]}) is None

        # stdlib types and corresponding serdes options
        stdlib_table.insert_one({"id": ALLMAPS_STDLIB_ROW["id"]})
        stdlib_table.update_one(
            {"id": ALLMAPS_STDLIB_ROW["id"]},
            update={"$set": {k: v for k, v in ALLMAPS_STDLIB_ROW.items() if k != "id"}},
        )
        assert (
            stdlib_table.find_one({"id": ALLMAPS_STDLIB_ROW["id"]})
            == ALLMAPS_STDLIB_ROW
        )
        stdlib_table.delete_one({"id": ALLMAPS_STDLIB_ROW["id"]})
        assert stdlib_table.find_one({"id": ALLMAPS_STDLIB_ROW["id"]}) is None

    @pytest.mark.describe("test of table maps-as-tuples, empty-map update one, sync")
    def test_table_mapsastuples_emptymaps_update_one_sync(
        self,
        sync_empty_table_allmaps: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_allmaps.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        customdt_table = tuplified_table
        stdlib_table = tuplified_table.with_options(api_options=STDLIB_OPTIONS)

        # custom types and corresponding serdes options
        tuplified_table.insert_one({"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]})
        tuplified_table.update_one(
            {"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]},
            update={
                "$set": {
                    k: v for k, v in ALLMAPS_CUSTOMTYPES_EMPTY_ROW.items() if k != "id"
                }
            },
        )
        tuplified_table.insert_one({"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]})
        tuplified_table.update_one(
            {"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]},
            update={
                "$set": {k: v for k, v in ALLMAPS_STDLIB_EMPTY_ROW.items() if k != "id"}
            },
        )

        # reading with a customtype and a stdlib-configured table:
        ct_e_ctrow = customdt_table.find_one(
            {"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        st_e_ctrow = customdt_table.find_one(
            {"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        dict_equal_same_class(ct_e_ctrow, st_e_ctrow)

        ct_e_strow = stdlib_table.find_one(
            {"id": ALLMAPS_CUSTOMTYPES_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        st_e_strow = stdlib_table.find_one(
            {"id": ALLMAPS_STDLIB_EMPTY_ROW["id"]},
            projection={"id": False},
        )
        dict_equal_same_class(ct_e_strow, st_e_strow)

    @pytest.mark.describe("test of table maps-as-tuples with ordinary rows, sync")
    def test_table_mapsastuples_ordinary_rows_sync(
        self,
        sync_empty_table_all_returns: DefaultTable,
    ) -> None:
        tuplified_table = sync_empty_table_all_returns.with_options(
            api_options=MAP2TUPLES_OPTIONS
        )
        FULL_AR_ROWS = DISTINCT_AR_ROWS + DISTINCT_EMPTYMAP_AR_ROWS
        tuplified_table.insert_many(FULL_AR_ROWS)
        assert len(tuplified_table.find({}).to_list()) == len(FULL_AR_ROWS)
